import ctypes
import os
import sys
import time
from collections import deque

from OpenGL import GL as gl
from OpenGL.GL.ARB.shader_objects import (
    glCompileShaderARB,
    glCreateShaderObjectARB,
    glDeleteObjectARB,
    glShaderSourceARB,
)

from arcade.gamecenter import GameCenter
from arcade.gamecentersettings import ArcadeSettings
from arcade.glui.animation import AnimateValueBezier, AnimationSystem
from arcade.glui.bezier import Bezier
from arcade.glui.constants import TOP_HEIGHT
from arcade.glui.displaylists import DisplayLists
from arcade.glui.font import BitmapFont, Font
from arcade.glui.imageloader import ImageLoader
from arcade.glui.input import InputHandler
from arcade.glui.itemmenu import ItemMenu
from arcade.glui.items import (
    AllMenuItem,
    ListItem,
    MenuItem,
    NoItem,
    PlatformItem,
)
from arcade.glui.notificationrender import NotificationRender
from arcade.glui.opengl import fs_emu_blending, fs_emu_texturing
from arcade.glui.render import Render
from arcade.glui.sdl import SDL_IsMinimized
from arcade.glui.settings import Settings
from arcade.glui.state import State
from arcade.glui.texture import Texture
from arcade.glui.texturemanager import TextureManager

# from arcade.main import Main
from arcade.resources import logger, resources
from fsbc import settings
from fsbc.system import windows
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.option import Option
from fsgs.platform import PlatformHandler
from fsgs.util.gamenameutil import GameNameUtil

# FIXME: rename to manager or director or somethign
main_window = None

# ENABLE_VSYNC = Config.get_bool("Menu/VSync", True)
ENABLE_VSYNC = "--vsync" in sys.argv
ALWAYS_RENDER = True
# IDLE = Config.get_bool("Menu/Idle", 1)
IDLE = 1
if "--no-idle" in sys.argv:
    IDLE = 0
# Render.get().display_fps = pyapp.user.ini.get_bool("Menu/ShowFPS", False)
# LIGHTING = Config.get_bool("Menu/Lighting", False)
LIGHTING = False
RENDER_DEBUG_SQUARES = 0
if "--render-debug-squares" in sys.argv:
    RENDER_DEBUG_SQUARES = 1
SEARCH_CHARS = (
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789 -:.,"
)
CONFIG_SEPARATION = 0.15
ANIMATION_SETTLE_TIME_1 = 1.0
ANIMATION_SETTLE_TIME_2 = 2.0

force_display_ratio = None
display = None
real_display_height = 0
current_menu = None
last_menu = None
last_game = None


def get_current_time():
    return time.time()


def set_current_menu(menu):
    if not isinstance(menu, SearchResultsMenu):
        menu.search_text = ""

    global current_menu
    if len(menu) == 0:
        menu.append(NoItem())
    current_menu = menu
    State.get().current_menu = menu
    if menu.navigatable:
        State.get().navigatable = menu.navigatable
    else:
        State.get().navigatable = menu
    State.get().top_menu = menu.top
    State.get().down_navigatable = None
    Render.get().dirty = True


# WINDOWED_SIZE = [1024, 640]  # 16:10
# WINDOWED_SIZE = (1024, 576)  # 16:9
# WINDOWED_SIZE = [1024, 768]  # 4:3
# WINDOWED_SIZE = [960, 540]  # 16:9
WINDOWED_SIZE = [1280, 720]  # 16:9

RENDER_GAME = ["IMAGE", "REFLECTIONS", "SCREENSHOTS"]
RENDER_GAME_OVERLAY = ["TITLE", "HEADER"]

# USE_MENU_TRANSITIONS = pyapp.user.ini.get_bool("Menu/Transitions", True)
# if "--disable-shaders" in sys.argv:
#     USE_MENU_TRANSITIONS = False
USE_MENU_TRANSITIONS = False
FIELD_COLOR = 0.1
WALL_COLOR = 0.0
# IDLE_EVENT = pygame.NUMEVENTS - 1
DEFAULT_ITEM_BRIGHTNESS = 0.8


class MenuGameTransition(object):
    value = 0.0


class MenuForwardTransition(object):
    # must be initially set to 1 because this indicates that
    # the transition is done / not in progress
    value = 1.0


class MenuBackwardTransition(object):
    # must be initially set to 1 because this indicates that
    # the transition is done / not in progress
    value = 1.0


class ScreenshotTransition(object):
    anim = None
    # must be initially set to 1 because this indicates that
    # the transition is done / not in progress
    value = 1.0


class RunTransition(object):
    value = 0.0


class Mouse(object):
    items = []
    focus = None

    @classmethod
    def set_visible(cls, visible=True):
        assert visible is not None
        # if State.get().mouse_visible == visible:
        #     return
        # if not visible:
        #     cls.focus = None
        # pygame.mouse.set_visible(visible)
        # State.get().mouse_visible = visible
        print("Mouse.set_visible not implemented")


def set_items_brightness(brightness, duration=1.0, delay=0.0):
    State.get().items_brightness_anim = AnimateValueBezier(
        (State, "items_brightness"),
        State.get().items_brightness,
        State.get().time + delay,
        State.get().items_brightness,
        State.get().time + delay,
        brightness,
        State.get().time + delay + duration,
        brightness,
        State.get().time + delay + duration,
    )


def compile_shader(source, shader_type):
    shader = glCreateShaderObjectARB(shader_type)
    glShaderSourceARB(shader, source)
    glCompileShaderARB(shader)
    try:
        status = ctypes.c_int()
        gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS, ctypes.byref(status))
        status = status.value
    except TypeError:
        status = gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS)
    if not status:
        print_log(shader)
        glDeleteObjectARB(shader)
        raise ValueError("Shader compilation failed")
    return shader


def compile_program(vertex_source, fragment_source):
    vertex_shader = None
    fragment_shader = None
    program = gl.glCreateProgram()
    if vertex_source:
        print("compile vertex shader")
        vertex_shader = compile_shader(vertex_source, gl.GL_VERTEX_SHADER)
        gl.glAttachShader(program, vertex_shader)
    if fragment_source:
        print("compile fragment shader")
        fragment_shader = compile_shader(
            fragment_source, gl.GL_FRAGMENT_SHADER
        )
        gl.glAttachShader(program, fragment_shader)
    gl.glLinkProgram(program)
    status = gl.glGetProgramiv(program, gl.GL_LINK_STATUS)
    if status == gl.GL_FALSE:
        print("could not link shader program")
        print(gl.glGetProgramInfoLog(program))
        sys.exit(1)
    if vertex_shader:
        gl.glDeleteShader(vertex_shader)
    if fragment_shader:
        gl.glDeleteShader(fragment_shader)
    return program


def print_log(shader):
    length = ctypes.c_int()
    gl.glGetShaderiv(shader, gl.gl.GL_INFO_LOG_LENGTH, ctypes.byref(length))
    if length.value > 0:
        shader_log = gl.glGetShaderInfoLog(shader)
        print(shader_log)
        sys.exit(1)


texture_program = None
color_program = None
premultiplied_texture_program = None


def compile_programs():
    global texture_program, color_program, premultiplied_texture_program
    vertex_shader = None

    color_program = compile_program(
        vertex_shader,
        [
            """
void main()
{
vec4 s = gl_Color;
float extra_alpha = 1.0;
//float extra_alpha = gl_TextureEnvColor[0].a;
float a = s.a * extra_alpha;
gl_FragColor.a = a;
gl_FragColor.r = s.r*a;
gl_FragColor.g = s.g*a;
gl_FragColor.b = s.b*a;
}
            """
        ],
    )

    texture_program = compile_program(
        vertex_shader,
        [
            """
uniform sampler2D texture;

void main()
{
vec4 s = texture2D(texture,gl_TexCoord[0].st);
//float extra_alpha = 1.0;
//float extra_alpha = gl_TextureEnvColor[0].a;
float opacity = gl_Color.a;
float a = s.a * opacity;
gl_FragColor.a = a;
gl_FragColor.r = s.r * a;
gl_FragColor.g = s.g * a;
gl_FragColor.b = s.b * a;
}
"""
        ],
    )

    premultiplied_texture_program = compile_program(
        vertex_shader,
        [
            """
uniform sampler2D texture;

void main()
{
vec4 s = texture2D(texture,gl_TexCoord[0].st);
//float extra_alpha = 1.0;
//float extra_alpha = gl_TextureEnvColor[0].a;
float opacity = gl_Color.a;
gl_FragColor.a = s.a * opacity;
gl_FragColor.r = s.r * opacity;
gl_FragColor.g = s.g * opacity;
gl_FragColor.b = s.b * opacity;
}
"""
        ],
    )


def enable_texture_shader():
    fs_emu_blending(True)
    gl.glBlendFunc(gl.GL_ONE, gl.GL_ONE_MINUS_SRC_ALPHA)
    gl.glUseProgram(texture_program)


def disable_shader():
    gl.glUseProgram(0)


def set_program(program):
    if not program:
        gl.glUseProgram(0)
    else:
        gl.glUseProgram(program)


def enter_menu(result, replace=False):
    print("enter_menu", result, "replace", replace)
    print("   menu parent_menu", result.parent_menu)

    if replace:
        State.get().history.pop()
    c_menu = State.get().history[-1]
    result.parents[:] = c_menu.parents
    result.parents.append(c_menu)
    print("   menu parents    ", result.parents)
    State.get().history.append(result)
    set_current_menu(result)


def create_main_menu():
    new_menu = AllMenuItem().activate(None)
    if len(new_menu) == 0:
        new_menu.append(NoItem())
    return new_menu


def recreate_main_menu_if_necessary():
    new_menu = create_main_menu()
    State.get().history.append(new_menu)
    set_current_menu(new_menu)


def show():
    global current_menu

    # fade_from is used on init_display, so we initialize this
    # color here. Set alpha to 2.0 to force 1 second of solid
    # color in combination with 2 sec. animation below
    if False and windows and not Settings.fullscreen_menu:
        State.get().fade_from = (1.0, 1.0, 1.0, 2.0)
        State.get().fade_to = (1.0, 1.0, 1.0, 0.0)
    else:
        State.get().fade_from = (0.0, 0.0, 0.0, 2.0)
        State.get().fade_to = (0.0, 0.0, 0.0, 0.0)
    init_display()
    if LIGHTING:
        init_lighting()
    init_textures()
    init_fonts()

    if USE_MENU_TRANSITIONS:
        compile_programs()

    InputHandler.open()

    on_resize((Render.get().display_width, Render.get().display_height))
    image_loader = ImageLoader.get()
    image_loader.start()
    new_menu = create_main_menu()
    State.get().history.append(new_menu)

    override_items = []
    for platform_id in PlatformHandler.get_platform_ids():
        if "--platform=" + platform_id in sys.argv:
            sys.argv.remove("--platform=" + platform_id)
            platform_item = PlatformItem(platform_id)
            override_items.append(platform_item)
            break
    # if "--favorites" in sys.argv or \
    #         settings.get(Option.ARCADE_INITIAL_VIEW) == "favorites":
    if (
        "--favorites" in sys.argv
        or settings.get(Option.ARCADE_INITIAL_FAVORITES) == "1"
    ):
        if "--favorites" in sys.argv:
            sys.argv.remove("--favorites")
        try:
            list_uuid = ListItem.get_favorites_uuid()
            favorites_item = ListItem("Favorites", list_uuid)
        except LookupError:
            print("No favorites list")
            # noinspection SpellCheckingInspection
            favorites_item = ListItem(
                "Favorites", "c03dd5fe-0e85-4efb-a126-f0e4f40feae6"
            )
        override_items.append(favorites_item)
    if override_items:
        parent_menu = new_menu
        override_menu = None
        parents = []
        for item in override_items:
            override_menu = ItemMenu()
            override_menu.parent_menu = parent_menu
            parents.append(parent_menu)
            # This is really hackish
            override_menu.parents = parents
            override_menu.items.append(item)
            parent_menu = override_menu
        new_menu = override_items[-1].activate(override_menu)
        print(new_menu)
        State.get().history.append(new_menu)

    set_current_menu(new_menu)
    if len(new_menu) == 1:
        # only n/a item showing, possibly
        if len(AllMenuItem().activate(None)) == 0:
            # no games, initiate game scan
            rescan_games()

    State.get().fade_start = get_current_time()
    State.get().fade_end = get_current_time() + 2.000

    # # make a timer so that update events are sent to modules at least once
    # # every second
    # pygame.time.set_timer(IDLE_EVENT, 1000)

    State.get().start_time = get_current_time()


def init_fonts():
    NotificationRender.init()

    BitmapFont.title_font = BitmapFont("title_font")
    BitmapFont.menu_font = BitmapFont("menu_font")


liberation_sans_bold_path = None
liberation_sans_narrow_bold_path = None
vera_font_path = None


def find_font_path_or_stream(name):
    path = os.path.join(
        FSGSDirectories.get_base_dir(), "Workspace", "Fonts", name
    )
    if os.path.exists(path):
        return path
    stream = resources.resource_stream(name)
    return stream


def reinit_fonts():
    print(
        "reinit_fonts, Render.get().display_height =",
        Render.get().display_height,
    )
    global liberation_sans_bold_path
    global liberation_sans_narrow_bold_path
    global vera_font_path

    if liberation_sans_bold_path is None:
        liberation_sans_bold_path = find_font_path_or_stream(
            "LiberationSans-Bold.ttf"
        )

    # if liberation_sans_narrow_bold_path is None:
    #     liberation_sans_narrow_bold_path = resources.resource_filename(
    #         "LiberationSansNarrow-Bold.ttf")
    # if vera_font_path is None:
    #     vera_font_path = resources.resource_filename(
    #         "VeraBd.ttf")

    if Font.title_font is None:
        Font.title_font = Font(
            liberation_sans_bold_path, int(0.04 * Render.get().display_height)
        )
    Font.title_font.set_size(int(0.04 * Render.get().display_height))
    if Font.subtitle_font is None:
        Font.subtitle_font = Font(
            liberation_sans_bold_path, int(0.025 * Render.get().display_height)
        )
    Font.subtitle_font.set_size(int(0.025 * Render.get().display_height))
    # if Font.small_font is None:
    #     Font.small_font = Font(
    #         liberation_sans_narrow_bold_path,
    #         int(0.025 * Render.get().display_height))
    # Font.small_font.set_size(int(0.025 * Render.get().display_height))
    if Font.main_path_font is None:
        Font.main_path_font = Font(
            liberation_sans_bold_path, int(0.025 * Render.get().display_height)
        )
    Font.main_path_font.set_size(int(0.025 * Render.get().display_height))
    if Font.list_subtitle_font is None:
        Font.list_subtitle_font = Font(
            liberation_sans_bold_path, int(0.020 * Render.get().display_height)
        )
    Font.list_subtitle_font.set_size(int(0.020 * Render.get().display_height))
    # if Font.header_font is None:
    #     Font.header_font = Font(
    #         vera_font_path, int(0.06 * Render.get().display_height))
    # Font.header_font.set_size(int(0.06 * Render.get().display_height))

    if NotificationRender.font is None:
        NotificationRender.font = Font(
            liberation_sans_bold_path, int(0.020 * Render.get().display_height)
        )


char_buffer = ""
char_buffer_last_updated = 0


def handle_search_menu_on_character_press(char_buffer):
    # if len(char_buffer) > 2 and t - char_buffer_last_updated > 0.5:
    if len(char_buffer) > 2:
        print(char_buffer)
        create_search_results_menu(char_buffer)
    if len(char_buffer) <= 2:
        if isinstance(current_menu, SearchResultsMenu):
            # collapse search menu

            # if hasattr(current_menu, "search_text"):
            print("setting current_menu (something with search)")
            # set_current_menu(current_menu.parent_menu)
            go_back()
        else:
            current_menu.search_text = char_buffer


# noinspection PyUnusedLocal
def character_press(char):
    print("character press", repr(char))
    if State.get().current_game:  # or isinstance(current_menu, GameMenu):
        print("ignoring key press", repr(char))
        return

    # global char_buffer
    char_buffer = current_menu.search_text

    global char_buffer_last_updated
    char_buffer_last_updated = State.get().time
    # return

    Render.get().dirty = True
    if char == "RETURN":
        print(char_buffer, len(char_buffer))
        print("returning false")
        return False
    elif char == "BACKSPACE" or char == "\b":
        if ArcadeSettings().search():
            char_buffer = char_buffer[:-1]
            if len(char_buffer) <= 2:
                char_buffer = ""
            handle_search_menu_on_character_press(char_buffer)
        return True
    elif len(char) == 1:
        if ArcadeSettings().search():
            char_buffer += char
            handle_search_menu_on_character_press(char_buffer)
            if 1 <= len(char_buffer) <= 2:
                jump_to_item = -1
                for i, item in enumerate(current_menu.items):
                    # FIXME: a bit hack-y this, should check sort_name
                    # instead (need to store this in items then)
                    # also, should use binary search and not sequential search
                    searches = [
                        char_buffer,
                        "the " + char_buffer,
                        "a " + char_buffer,
                    ]
                    check = item.name.lower()
                    for s in searches:
                        if check.startswith(s):
                            jump_to_item = i
                            break
                if jump_to_item >= 0:
                    current_menu.set_selected_index(
                        jump_to_item, immediate=True
                    )
        return True
    else:
        raise Exception(repr(char))


class SearchTextItem(MenuItem):
    def __init__(self, title):
        MenuItem.__init__(self)
        self.path_title = title


class SearchResultsMenu(ItemMenu):
    pass


def create_search_results_menu(text):
    global current_menu
    try:
        if text == current_menu.search_text:
            return False
    except AttributeError:
        pass
    new_menu = SearchResultsMenu("Search Results")
    new_menu.search_text = text
    # words = [v.strip() for v in text.lower().split(" ")]
    # print "Creating search results for", words
    new_menu.top.append_left(SearchTextItem("Search: {0}_".format(text)))
    new_menu.top.set_selected_index(
        len(new_menu.top.left) + len(new_menu.top.right) - 1
    )

    # clause = []
    # args = []
    # for word in words:
    #     clause.append("AND name like ?")
    #     args.append("%{0}%".format(word))
    # clause = " ".join(clause)
    terms = GameNameUtil.extract_search_terms(text.lower())
    for item in MenuItem.create_game_items(terms):
        new_menu.append(item)
    if len(new_menu) == 0:
        new_menu.append(NoItem("No Search Results"))
    # if hasattr(current_menu, "search_text"):
    #     # replace current search menu, not append to path
    #     #new_menu.parent_menu = current_menu.parent_menu
    #     replace = True
    # else:
    #     #new_menu.parent_menu = current_menu
    #     replace = False
    replace = isinstance(current_menu, SearchResultsMenu)
    print("create search results menu")
    # set_current_menu(new_menu)
    enter_menu(new_menu, replace=replace)
    return True


def rescan_games():
    # global current_menu
    print("rescan games -- currently disabled")
    # GameScanner.scan()
    # Render.get().dirty = True
    pass


def go_back():
    c_menu = State.get().history[-1]
    if hasattr(c_menu, "go_back"):
        c_menu.go_back()
        return
    State.get().history.pop()
    set_current_menu(State.get().history[-1])


def default_input_func(button):
    if button == "LEFT":
        State.get().navigatable.go_left()
    elif button == "RIGHT":
        State.get().navigatable.go_right()
    elif button == "UP":
        State.get().navigatable.go_up()
    elif button == "DOWN":
        State.get().navigatable.go_down()
    elif button == "SKIP_LEFT":
        State.get().navigatable.go_left(10)
    elif button == "SKIP_RIGHT":
        State.get().navigatable.go_right(10)
    elif button == "PRIMARY":
        # global char_buffer
        # char_buffer = ""

        State.get().navigatable.activate()

    elif button == "BACK":
        print("-- button is BACK -- ")
        if State.get().config_menu:
            State.get().config_menu = None
            set_current_menu(State.get().current_menu)
            set_items_brightness(0.66, duration=0.500)
        elif State.get().current_game is None and current_menu.search_text:
            InputHandler.get_button()  # clear OK status
            character_press("BACKSPACE")
        elif State.get().current_game:
            State.get().current_game = None
            # game_fade_animation = AnimateValueBezier(
            #     (MenuGameTransition, "value"),
            #     1.0, State.get().time,
            #     1.0, State.get().time + 0.133,
            #     0.0, State.get().time + 0.133,
            #     0.0, State.get().time + 0.400)

        # elif can_navigate and current_menu.parent_menu:
        # elif can_navigate and len(State.get().history) > 1:
        elif len(State.get().history) > 1:
            go_back()
    elif button == "QUIT":
        State.get().quit = True


def render_top():
    Render.get().hd_perspective()
    gl.glPushMatrix()
    transition = State.get().current_menu.top_menu_transition
    gl.glTranslate(0.0, (1.0 - transition) * 90, 0.9)

    gl.glTranslate(0.0, 0.0, 0.05)

    if State.get().top_menu == State.get().navigatable:
        selected_index = State.get().top_menu.get_selected_index()
    else:
        selected_index = -1
    x = State.get().gl_left
    for item in State.get().top_menu.left:
        item.update_size_left()
        item.x = x
        item.y = 1080 - TOP_HEIGHT
        item.h = TOP_HEIGHT
        x += item.w

    index = len(State.get().top_menu.left) - 1
    for item in reversed(State.get().top_menu.left):
        item.render_top_left(selected=(index == selected_index))
        index -= 1

    index = len(State.get().top_menu) - 1
    x = State.get().gl_right
    for item in reversed(State.get().top_menu.right):
        item.update_size_right()
        x -= item.w
        item.x = x
        item.y = 1080 - TOP_HEIGHT
        item.h = TOP_HEIGHT
        if Mouse.focus:
            selected = Mouse.focus == item
        else:
            selected = index == selected_index
        item.render_top_right(selected=selected)
        Mouse.items.append(item)
        index -= 1
    gl.glPopMatrix()


def render_config_menu():
    if not State.get().config_menu:
        return

    Render.get().ortho_perspective()
    config_menu = State.get().config_menu
    # text = config_menu.items[config_menu.index]
    # otw, th = Render.get().text(text, title_font,
    #         -1.0, -0.93, 2.0, color=(1.0, 1.0, 1.0, 0.36 * strength))
    # x = 0.0 + otw / 2 + CONFIG_SEPARATION
    # for i in range(config_menu.index + 1, len(config_menu.items)):
    #     text = config_menu.items[i]
    #     tw, th = Render.get().text(text, title_font,
    #             x, -0.93, color=(1.0, 1.0, 1.0, 0.36 * strength))
    #     x += tw + CONFIG_SEPARATION
    # x = 0.0 - otw / 2 - CONFIG_SEPARATION
    x = -0.55
    y = 0.8
    for i in range(len(config_menu.items)):
        text = config_menu.items[i].upper()
        # tw, th = Render.get().measure_text(text, title_font)
        # x -= tw + CONFIG_SEPARATION
        y -= 0.15
        if i == config_menu.index and config_menu == State.get().navigatable:
            color = (1.0, 1.0, 1.0, 1.0)
        else:
            color = (1.0, 1.0, 1.0, 0.33)
        Render.get().text(text, Font.subtitle_font, x, y, color=color)


def render_scanning_status():
    # Render.get().hd_perspective()
    # text = GameScanner.get_status()
    # Render.get().dirty = True
    #
    # fs_emu_texturing(False)
    #
    # z = 0.0
    #
    # glBegin(GL_QUADS)
    # glColor3f(0.0, 0.0, 0.0)
    # glVertex3f(   0, 500, z)
    # glVertex3f(1920, 500, z)
    # glVertex3f(1920, 700, z)
    # glVertex3f(   0, 700, z)
    # glEnd()
    #
    # Render.get().text(text, Font.title_font, 200, 600, color=(1.0, 1.0,
    # 1.0, 1.0))
    #
    # glBegin(GL_QUADS)
    # glColor3f(1.0, 1.0, 1.0)
    # x = 200
    # y = 500
    # z = 0.9
    # x2 = 200 + 1520 * GameScanner.progress
    # glVertex3f(x, y, z)
    # glVertex3f(x2, y, z)
    # glVertex3f(x2, y + 20, z)
    # glVertex3f(x, y + 20, z)
    # glEnd()
    # fs_emu_texturing(True)
    pass


def do_render():
    if current_menu is None:
        return

    current_menu.update()

    if RunTransition.value > 0.99:
        # do not render anything when running a game
        return

    if State.get().currently_ingame:
        # print("currently ingame")
        return

    # try to exploit parallelism by uploading texture while rendering
    TextureManager.get().load_textures(1)

    # clear mouseover rects -these will be calculated during rendering
    Mouse.items[:] = []
    Render.get().standard_perspective()

    # scanning = GameScanner.scanning

    data = current_menu.render()
    current_menu.render_transparent(data)

    # if GameScanner.is_scanning():
    if False:
        render_scanning_status()
        State.get().was_scanning = True
    else:
        render_config_menu()
        if State.get().was_scanning:
            print("State.get().was_scanning")
            State.get().was_scanning = False
            # reload current menu

            # if current_menu.parent_menu:
            #     result = current_menu.parent_menu.selected_item.activate(
            #             current_menu.parent_menu)
            #     if isinstance(result, Menu):
            #         #if len(result) == 0:
            #         #    result.append(NoItem())
            #         result.parent_menu = current_menu.parent_menu
            #         print("set new menu (rescanned games)")
            #         set_current_menu(result)
            recreate_main_menu_if_necessary()

    render_top()

    if State.get().dialog:
        State.get().dialog.render()

    render_global_fade()
    NotificationRender.render()
    if RunTransition.value > 0.0:
        render_fade(a=RunTransition.value)


def render_fade(r=0.0, g=0.0, b=0.0, a=0.0):
    Render.get().hd_perspective()
    fs_emu_blending(True)
    fs_emu_texturing(False)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor4f(r * a, g * a, b * a, a)
    gl.glVertex2f(0, 0)
    gl.glVertex2f(1920, 0)
    gl.glVertex2f(1920, 1080)
    gl.glVertex2f(0, 1080)
    gl.glEnd()


def render_global_fade():
    t = State.get().time
    if State.get().fade_end >= t >= State.get().fade_start:
        a = (t - State.get().fade_start) / (
            State.get().fade_end - State.get().fade_start
        )
        if a < 0.0:
            a = 0.0
        elif a > 1.0:
            a = 1.0

        Render.get().hd_perspective()
        gl.glPushMatrix()
        gl.glTranslatef(0.0, 0.0, 0.99999)
        fs_emu_blending(True)
        fs_emu_texturing(True)
        if State.get().fade_splash:
            Texture.splash.render(
                (1920 - Texture.splash.w) // 2,
                (1080 - Texture.splash.h) // 2,
                Texture.splash.w,
                Texture.splash.h,
                opacity=(1.0 - a),
            )

        c = [0, 0, 0, (1.0 - a)]
        # for i in range(4):
        #     c[i] = State.get().fade_from[i] + \
        #             (State.get().fade_to[i] - State.get().fade_from[i]) * a
        #  * (a)
        render_fade(*c)
        gl.glPopMatrix()
        Render.get().dirty = True


FPS_FRAMES = 100
render_times = deque()
for _ in range(FPS_FRAMES):
    render_times.append(0)

fps_str = ""
debug_x = 0
debug_x_2 = 0


def render_debug_square():
    global debug_x
    Render.get().hd_perspective()
    fs_emu_texturing(False)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor3f(1.0, 1.0, 1.0)
    x = debug_x
    debug_x += 1
    if debug_x >= 1920:
        debug_x = 0
    y = 989
    z = 0.99
    gl.glVertex3f(x, y, z)
    gl.glVertex3f(x + 20, y, z)
    gl.glVertex3f(x + 20, y + 20, z)
    gl.glVertex3f(x, y + 20, z)
    gl.glEnd()
    fs_emu_texturing(True)


def render_debug_square_2():
    global debug_x_2
    Render.get().hd_perspective()
    fs_emu_texturing(False)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor3f(0.2, 0.2, 0.2)
    x = debug_x_2
    debug_x_2 += 1
    if debug_x_2 >= 1920:
        debug_x_2 = 0
    y = 989 + 5
    z = 0.99
    gl.glVertex3f(x, y, z)
    gl.glVertex3f(x + 20, y, z)
    gl.glVertex3f(x + 20, y + 10, z)
    gl.glVertex3f(x, y + 10, z)
    gl.glEnd()
    fs_emu_texturing(True)


# FIXME: remove / move some code away
def swap_buffers():
    # global fps_str
    #
    # Render.get().ortho_perspective()
    # #glPushMatrix()
    # #glTranslatef(0.0, 0.0, 0.5)
    #
    # #if not fps_str:
    # #    fps_str = "WHY?"
    #
    # if Render.get().display_fps or True:
    #     if fps_str:
    #         Render.get().text(fps_str, Font.main_path_font,
    #                     -1.74, 0.82, h=0.1, color=(0.25, 0.25, 0.25, 1.0))
    # #glPopMatrix()
    #
    # # FIXME: Why does not minimize from fullscreen work on ATI unless we
    # render
    # # something here?
    # glBindTexture(GL_TEXTURE_2D, 0)
    # glPushMatrix()
    # glTranslate(2000.0, 0.0, 0.0)
    # glBegin(GL_QUADS)
    # glVertex2f(0.0, 0.0)
    # glVertex2f(1.0, 0.0)
    # glVertex2f(1.0, 1.0)
    # glVertex2f(0.0, 1.0)
    # glEnd()
    # glPopMatrix()
    #
    # #fs_emu_blending(False)
    # #glEnable(GL_DEPTH_TEST)
    #
    # #if Render.get().display_sync:
    # #    glFinish()
    #
    # #pygame.display.flip()
    # print("FIXME: not flipping")
    #
    # if Render.get().display_sync:
    #     # give up time slice
    #     time.sleep(0.001)
    #     glFinish()
    #
    # if Render.get().display_fps:
    #     t = get_current_time()
    #     render_times.append(t)
    #     t0 = render_times.popleft()
    #     if t0 > 0 and State.get().frame_number % 5 == 0:
    #         time_diff = t - t0
    #         #print("{0:0.2f}".format(300.0 / time_diff))
    #         fps = FPS_FRAMES / time_diff
    #         if fps >= 100:
    #             fps_str = "FPS:  {0:0.0f}".format(fps)
    #         else:
    #             fps_str = "FPS: {0:0.1f}".format(fps)
    #         #Render.get().text(fps_str, Font.title_font,
    #         #        -1.0, 0.90, 2.0, shadow=True)
    State.get().frame_number += 1
    Render.get().delete_textures()


def render_screen():
    # set Render.get().dirty to False here, so that render functions can
    # request a new render frame by setting dirty
    Render.get().dirty = False
    # glEnable(GL_SCISSOR_TEST)
    # can_idle =
    if SDL_IsMinimized():
        time.sleep(0.01)
        return
    # Render.get().dirty = True

    error = gl.glGetError()
    if error:
        # FIXME: Why error?
        print(error)

    gl.glClearColor(0.0, 0.0, 0.0, 1.0)
    gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
    do_render()
    if Render.get().display_fps:
        Render.get().dirty = True
    if RENDER_DEBUG_SQUARES:
        render_debug_square()


# noinspection PyPep8Naming
def init_display():
    global display
    global real_display_height  # , display_yoffset

    # global banner_texture, shadow_texture, gloss_texture
    # global top_texture, top_logo_texture, logo_texture
    # global missing_cover_texture, default_item_texture
    # global backdrop_texture

    logger.debug("Init OpenGL menu display")

    DisplayLists.clear()

    # on_resize()
    # depth = 0
    # FIXME: HACK / TESTING
    # if not Settings.fullscreen_menu:
    #    if windows:
    #        os.environ["SDL_VIDEO_WINDOW_POS"] = "3,29"
    #    else:
    #        os.environ["SDL_VIDEO_WINDOW_POS"] = "0,0"
    # maximize_window = (not Settings.fullscreen_menu and
    #                    Settings.windowed_size is None)

    # display_info = pygame.display.Info()
    # dw = display_info.current_w
    # dh = display_info.current_h
    # dw, dh = fsui.get_screen_size()
    # dw, dh = 100, 100
    Render.get().display_width = main_window.width
    Render.get().display_height = main_window.height

    # if Settings.fullscreen_menu:
    #     print("fullscreen is True")
    #     if windows:
    #         #resolution = (0, 0)
    #         #flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.NOFRAME \
    #         #        | pygame.FULLSCREEN
    #         os.environ["SDL_VIDEO_WINDOW_POS"] = "0,0"
    #         flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.NOFRAME
    #         #flags = flags | pygame.FULLSCREEN
    #         #if fs.linux:
    #         #    pass
    #         #else:
    #         if dw > dh * 2:
    #             # Assume dual monitor setup - hack for Linux / SDL
    #             resolution = (dw / 2, dh)
    #         else:
    #             resolution = (dw, dh)
    #     else:  # fullscreen, but not microsoft windows
    #         #resolution = (0, 0)
    #         flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.NOFRAME #|
    # pygame.FULLSCREEN
    #         if dw > dh * 2:
    #             # Assume dual monitor setup - hack for Linux / SDL
    #             resolution = (dw / 2, dh)
    #         else:
    #             resolution = (dw, dh)
    #         if linux:
    #             overscan = Config.get("display/overscan", "0,0,0,0")
    #             try:
    #                 overscan = overscan.split(",")
    #                 overscan = [int(x.strip()) for x in overscan]
    #                 print("using overscan", overscan)
    #             except Exception as e:
    #                 print("error parsing overscan from config:", repr(e))
    #                 overscan = [0, 0, 0, 0]
    #             os.environ["SDL_VIDEO_WINDOW_POS"] = "{0},{1}".format(
    #                 overscan[0], overscan[1])
    #             resolution = (resolution[0] - overscan[0] - overscan[2],
    #                           resolution[1] - overscan[1] - overscan[3])
    #         elif macosx:
    #             # FIXME: fullscreen mode does not work very well. -When
    #             # opening a fullscreen emulator from fullscreen, the emulator
    #             # crashes on glViewport. Tested with fs-amiga.
    #             #flags |= pygame.FULLSCREEN
    #
    #             # for now, we create an almost maximized window, works
    #             # quite well when the dock is set to auto-hide
    #             #resolution = (resolution[0], resolution[1] - 22)
    #
    #             # FIXME: trying LSUIPresentationMode
    #             os.environ["SDL_VIDEO_WINDOW_POS"] = "0,0"
    #
    #             # kUIModeNormal = 0
    #             # kUIModeContentSuppressed = 1
    #             # kUIModeContentHidden = 2
    #             # kUIModeAllSuppressed = 4
    #             kUIModeAllHidden = 3
    #             kUIOptionAutoShowMenuBar = 1 << 0
    #             # kUIOptionDisableAppleMenu = 1 << 2
    #             # kUIOptionDisableProcessSwitch = 1 << 3
    #             # kUIOptionDisableForceQuit = 1 << 4
    #             # kUIOptionDisableSessionTerminate = 1 << 5
    #             # kUIOptionDisableHide = 1 << 6
    #
    #             #noinspection PyUnresolvedReferences
    #             import objc
    #             #noinspection PyUnresolvedReferences
    #             from Foundation import NSBundle
    #             bundle = NSBundle.bundleWithPath_(
    #                 "/System/Library/Frameworks/Carbon.framework")
    #             objc.loadBundleFunctions(
    #                 bundle, globals(),
    #                 ((str("SetSystemUIMode"), str("III"), str("")),))
    #             #noinspection PyUnresolvedReferences
    #             SetSystemUIMode(kUIModeAllHidden, kUIOptionAutoShowMenuBar)
    #
    # else:
    #     if windows and maximize_window and \
    #             not Settings.window_decorations:
    #         import ctypes
    #         SPI_GETWORKAREA = 48
    #
    #         class RECT(ctypes.Structure):
    #             _fields_ = [
    #                 ("left", ctypes.c_ulong),
    #                 ("top", ctypes.c_ulong),
    #                 ("right", ctypes.c_ulong),
    #                 ("bottom", ctypes.c_ulong)]
    #
    #         m = ctypes.windll.user32
    #         r = RECT()
    #         m.SystemParametersInfoA(SPI_GETWORKAREA, 0, ctypes.byref(r), 0)
    #         x = int(r.left)
    #         y = int(r.top)
    #         w = int(r.right) - int(r.left)
    #         h = int(r.bottom) - int(r.top)
    #         print(x, y, w, h)
    #         WINDOWED_SIZE[0] = w
    #         WINDOWED_SIZE[1] = h
    #         os.environ["SDL_VIDEO_WINDOW_POS"] = "{0},{1}".format(x, y)
    #         State.get().allow_minimize = False
    #
    #     if Settings.windowed_size:
    #         print("Settings.windowed_size", Settings.windowed_size)
    #         WINDOWED_SIZE[0] = Settings.windowed_size[0]
    #         WINDOWED_SIZE[1] = Settings.windowed_size[1]
    #         Render.get().display_width = WINDOWED_SIZE[0]
    #         Render.get().display_height = WINDOWED_SIZE[1]
    #         #if dw > 1400:
    #         #    Render.get().display_width = 1280
    #         #    Render.get().display_height = 720
    #     else:
    #         Render.get().display_width = WINDOWED_SIZE[0]
    #         Render.get().display_height = WINDOWED_SIZE[1]
    #     resolution = (Render.get().display_width, Render.get().display_height)
    #     #print(resolution)
    #     #sys.exit(1)
    #     if Settings.window_decorations:
    #         flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.RESIZABLE
    #     else:
    #         flags = pygame.OPENGL | pygame.DOUBLEBUF | pygame.NOFRAME
    #
    # display_yoffset = 0
    # Mouse.set_visible(False)
    #
    # pygame.display.gl_set_attribute(pygame.GL_STENCIL_SIZE, 8)
    # pygame.display.gl_set_attribute(pygame.GL_DEPTH_SIZE, 16)
    #
    # Render.get().display_sync = ENABLE_VSYNC
    # if Render.get().display_sync:
    #     print("enabling vertical sync")
    #     os.environ["__GL_SYNC_TO_VBLANK"] = "1"
    #     pygame.display.gl_set_attribute(pygame.GL_SWAP_CONTROL, 1)
    # else:
    #     os.environ["__GL_SYNC_TO_VBLANK"] = "0"
    #     pygame.display.gl_set_attribute(pygame.GL_SWAP_CONTROL, 0)
    # pygame.display.gl_set_attribute(pygame.GL_DOUBLEBUFFER, 1)
    # fsaa = Config.get_int("video/fsaa", 0)
    # if fsaa:
    #     pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLEBUFFERS, 1)
    #     pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLESAMPLES, fsaa)
    # print("pygame set display mode", resolution, flags, depth)
    # display = pygame.display.set_mode(resolution, flags, depth)
    # if not Settings.fullscreen_game:
    #     try:
    #         del os.environ["SDL_VIDEO_WINDOW_POS"]
    #     except KeyError:
    #         pass
    #
    # if app.name == "fs-uae-arcade":
    #     pygame.display.set_caption("FS-UAE Arcade")
    # else:
    #     pygame.display.set_caption("FS Game Center")
    #
    # # FIXME: DISABLING MAXIMIZE FOR DEBUGGING
    # #maximize_window = False
    # if maximize_window:
    #     print("maximizing window")
    #     SDL_Maximize()
    #     for event in pygame.event.get():
    #         if event.type == pygame.VIDEORESIZE:
    #             #WINDOWED_SIZE[0] = event.w
    #             #WINDOWED_SIZE[1] = event.h
    #             on_resize((event.w, event.h))
    #     print("DISPLAY.GET_SIZE", display.get_size())
    # else:
    #     on_resize(display.get_size())

    gl.glMatrixMode(gl.GL_MODELVIEW)

    gl.glBlendFunc(gl.GL_ONE, gl.GL_ONE_MINUS_SRC_ALPHA)
    gl.glClearColor(*State.get().fade_from)

    fs_emu_texturing(True)
    Texture.splash = Texture("splash.png")

    for _ in range(0):
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        # gl.glClear(gl.GL_COLOR_BUFFER_BIT)
        Render.get().hd_perspective()
        Texture.splash.render(
            (1920 - Texture.splash.w) // 2,
            (1080 - Texture.splash.h) // 2,
            Texture.splash.w,
            Texture.splash.h,
        )
        gl.glFinish()
        # pygame.display.flip()
        gl.glFinish()

    gl.glEnable(gl.GL_DEPTH_TEST)


def init_textures():
    Texture.shadow = Texture.from_resource("shadow.png")
    Texture.shadow2 = Texture.from_resource("shadow2.png")
    # Texture.gloss = Texture.from_resource("gloss.png")
    Texture.gloss = Texture("gloss.png")
    Texture.screen_gloss = Texture("screen_gloss.png")
    Texture.static = Texture("preview_static0.png")
    Texture.default_item = Texture("default_item.png")
    Texture.missing_screenshot = Texture("missing_screenshot.png")
    Texture.missing_cover = Texture("missing_cover.png")
    # path = os.path.join(fs.get_data_dir(), "logo.png")
    # if os.path.exists(path):
    #     im = Image.open(path)
    #     Texture.logo = Texture.load(im)
    # else:
    Texture.logo = Texture.from_resource("logo.png")
    Texture.top = Texture.from_resource("top.png")
    Texture.top_logo = Texture("top_logo.png")
    Texture.top_logo_selected = Texture("top_logo_selected.png")

    Texture.add = Texture("add.png")
    Texture.add_selected = Texture("add_selected.png")
    Texture.home = Texture("home.png")
    Texture.home_selected = Texture("home_selected.png")
    Texture.minimize = Texture("minimize.png")
    Texture.minimize_selected = Texture("minimize_selected.png")
    Texture.close = Texture("close.png")
    Texture.close_selected = Texture("close_selected.png")
    Texture.shutdown = Texture("shutdown.png")
    Texture.shutdown_selected = Texture("shutdown_selected.png")

    Texture.bottom_bar = Texture("bottom_bar.png")
    Texture.screen_border_1 = Texture("screen_border_1.png")
    Texture.screen_border_2 = Texture("screen_border_2.png")
    Texture.top_background = Texture("top_background.png")
    Texture.top_item = Texture("top_item.png")
    Texture.top_item_selected = Texture("top_item_selected.png")
    Texture.top_item_left = Texture("top_item_left.png")
    Texture.top_item_left_selected = Texture("top_item_left_selected.png")
    Texture.top_item_right = Texture("top_item_right.png")
    Texture.top_item_arrow = Texture("top_item_arrow.png")
    Texture.top_item_arrow_selected = Texture("top_item_arrow_selected.png")

    Texture.sidebar_background = Texture("sidebar_background.png")
    Texture.sidebar_background_shadow = Texture(
        "sidebar_background_shadow.png"
    )
    Texture.glow_top = Texture("glow_top.png")
    Texture.glow_top_left = Texture("glow_top_left.png")
    Texture.glow_left = Texture("glow_left.png")

    Texture.heading_strip = Texture("heading_strip.png")
    Texture.item_background = Texture("item_background.png")
    Texture.top_item_background = Texture("top_item_background.png")
    Texture.logo_32 = Texture("logo-32.png")

    Texture.stretch = Texture("stretch.png")
    Texture.aspect = Texture("stretch-aspect.png")
    Texture.square_pixels = Texture("stretch-none.png")

    # # FIXME: TEMPORARY - FOR TESTING, ONLY
    # path = "c:\\git\\fs-game-database\\Backdrops\\ffx.png"
    # if os.path.exists(path):
    #     im = Image.open(path)
    #     #im = resources.get_resource_image_pil("shadow.png")
    #     assert im.size == (1024, 1024)
    #     imdata = im.tostring("raw", "RGB")
    #     backdrop_texture = Render.get().create_texture()
    #     glBindTexture(GL_TEXTURE_2D, backdrop_texture)
    #     glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, im.size[0], im.size[1], 0,
    #                  GL_RGB, GL_UNSIGNED_BYTE, imdata)
    #     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
    #     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
    # else:
    #     backdrop_texture = 0


def init_lighting():
    gl.glLightModeli(gl.GL_LIGHT_MODEL_LOCAL_VIEWER, gl.GL_TRUE)
    gl.glLightModeli(gl.GL_LIGHT_MODEL_TWO_SIDE, gl.GL_FALSE)
    gl.glLightModeli(
        gl.GL_LIGHT_MODEL_COLOR_CONTROL, gl.GL_SEPARATE_SPECULAR_COLOR
    )

    light_position = (0.0, 0.0, 3.0, 1.0)
    gl.glLightfv(gl.GL_LIGHT0, gl.GL_POSITION, light_position)
    gl.glLightfv(gl.GL_LIGHT0, gl.GL_DIFFUSE, (1.0, 1.0, 1.0, 1.0))
    gl.glLightfv(gl.GL_LIGHT0, gl.GL_SPECULAR, (0.0, 0.0, 0.0, 1.0))
    gl.glEnable(gl.GL_LIGHT0)

    gl.glLightfv(gl.GL_LIGHT1, gl.GL_DIFFUSE, (0.0, 0.0, 0.0, 1.0))
    gl.glLightfv(gl.GL_LIGHT1, gl.GL_SPECULAR, (1.0, 1.0, 1.0, 1.0))
    gl.glEnable(gl.GL_LIGHT1)

    gl.glLightfv(gl.GL_LIGHT2, gl.GL_DIFFUSE, (0.0, 0.0, 0.0, 1.0))
    gl.glLightfv(gl.GL_LIGHT2, gl.GL_SPECULAR, (0.5, 0.5, 0.5, 1.0))
    gl.glEnable(gl.GL_LIGHT2)

    gl.glMaterialfv(gl.GL_FRONT, gl.GL_AMBIENT, (0.1, 0.1, 0.1, 1.0))
    gl.glMaterialfv(gl.GL_FRONT, gl.GL_SHININESS, (10,))


def handle_videoresize_event(event):
    WINDOWED_SIZE[0] = event.w
    WINDOWED_SIZE[1] = event.h
    on_resize((event.w, event.h))


def on_resize(display_size):
    print("glui.window.on_resize", display_size)
    # traceback.print_stack()

    global display
    global real_display_height, display_yoffset
    global browse_curve, header_curve  # ,screenshot_curve

    error = gl.glGetError()
    if error:
        print(error)

    DisplayLists.clear()

    Render.get().display_width, Render.get().display_height = display_size  #
    #  display# .get_size()
    State.get().display_aspect = (
        Render.get().display_width / Render.get().display_height
    )
    print(WINDOWED_SIZE)
    Settings.windowed_size = tuple(WINDOWED_SIZE)

    real_display_height = Render.get().display_height
    # if Render.get().display_width / Render.get().display_height < 4 / 3:
    #    Render.get().display_height = int(Render.get().display_width / 4 * 3)
    #    display_yoffset = (real_display_height - Render.get(
    # ).display_height) // 2
    # else:
    display_yoffset = 0
    # glViewport(0, 0, Render.get().display_width, real_display_height)
    # glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    # swap_buffers()
    # glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    # swap_buffers()
    print(display_yoffset)

    reinit_fonts()

    print(
        0,
        display_yoffset,
        Render.get().display_width,
        Render.get().display_height,
    )
    gl.glViewport(
        0,
        display_yoffset,
        Render.get().display_width,
        Render.get().display_height,
    )

    # aspect_ratio = max(4 / 3, (Render.get().display_width / Render.get(
    # ).display_height))
    factor = (Render.get().display_width / Render.get().display_height) / (
        1024 / 600
    )
    browse_curve = Bezier.bezier(
        (-5.0 * factor, -10.0),
        (-1.7 * factor, -0.0),
        (1.7 * factor, -0.0),
        (5.0 * factor, -10.0),
    )
    header_curve = Bezier.bezier(
        (-2.0 * factor, 0.00),
        (-1.0 * factor, 0.075),
        (1.0 * factor, 0.075),
        (2.0 * factor, 0.00),
        steps=50,
    )

    # if USE_MENU_TRANSITIONS:
    #    State.get().fbo = FrameBufferObject(
    #        Render.get().display_width, Render.get().display_height)
    Render.get().dirty = True


class LogoStrength(object):
    anim = None
    value = 1.0


def set_ingame_status():
    # State.get().fade_start = State.get().time
    # State.get().fade_end = State.get().time + 86400.0 * 365.0
    # State.get().fade_from = (0.0, 0.0, 0.0, 0.0)
    # State.get().fade_to = (0.0, 0.0, 0.0, 0.0)
    # State.get().fade_splash = False
    print("State.get().currently_ingame = True")
    State.get().currently_ingame = True


def back_to_menu_from_game():
    print("State.get().currently_ingame = False")
    State.get().currently_ingame = False

    set_items_brightness(0.66, duration=0.500)

    RunTransition.value = 0.0
    RunTransition.anim = None
    LogoStrength.anim = None
    LogoStrength.value = 1.0

    Render.get().zoom = 1.0
    Render.get().offset_x = 0.0
    Render.get().offset_y = 0.0

    State.get().fade_splash = True
    State.get().fade_start = State.get().time
    State.get().fade_end = State.get().time + 0.5
    State.get().fade_from = (0.0, 0.0, 0.0, 2.0)
    State.get().fade_to = (0.0, 0.0, 0.0, 0.0)


def fade_quit():
    print("fade_quit")
    duration = 0.500
    alpha = 0.0
    start = get_current_time()
    while True:
        alpha = min(1.0, (get_current_time() - start) / duration)

        def render_func():
            render_screen()
            Render.get().ortho_perspective()
            fs_emu_blending(True)
            fs_emu_texturing(False)
            gl.glDisable(gl.GL_DEPTH_TEST)
            gl.glBegin(gl.GL_QUADS)
            gl.glColor4f(0.0, 0.0, 0.0, alpha)
            gl.glVertex2f(-10.0, -1.0)
            gl.glVertex2f(10.0, -1.0)
            gl.glVertex2f(10.0, 1.0)
            gl.glVertex2f(-10.0, 1.0)
            gl.glEnd()
            gl.glEnable(gl.GL_DEPTH_TEST)
            # fs_emu_blending(False)
            fs_emu_texturing(True)
            swap_buffers()
            Render.get().dirty = True

        Render.get().dirty = True
        main_loop_iteration(input_func=None, render_func=render_func)
        if alpha >= 1.0:
            break


def default_render_func():
    # rendered = False

    time_str = time.strftime("%H:%M")
    if time_str != Render.get().last_time_str:
        # clock must be updated, at least
        Render.get().dirty = True
        Render.get().last_time_str = time_str

    if Render.get().dirty or ALWAYS_RENDER:
        # print(Render.get().frame_number)
        render_screen()
        # rendered = True
        Render.get().twice = False
        # Render.get().twice = True
    else:
        if not Render.get().twice:
            Render.get().twice = True
            render_screen()
            # rendered = True
    if RENDER_DEBUG_SQUARES:
        render_debug_square_2()
    # if not rendered:
    #     pass
    # time.sleep(0.01)
    swap_buffers()


def find_item_at_coordinate(pos):
    menu = current_menu
    # Just checking top items for now
    for item in menu.top.left:
        if (
            item.x <= pos[0] <= item.x + item.w
            and item.y <= pos[1] <= item.y + item.h
        ):
            return item
    for item in menu.top.right:
        if (
            item.x <= pos[0] <= item.x + item.w
            and item.y <= pos[1] <= item.y + item.h
        ):
            return item
    return None


def handle_mouse_event(event):
    if event["type"] == "mouse-motion":
        item = find_item_at_coordinate(event["pos"])
        # print("mouse over item", item)
        state = State.get()
        state.mouse_item = item
    elif event["type"] == "mouse-press":
        item = find_item_at_coordinate(event["pos"])
        state = State.get()
        state.mouse_press_item = item
    elif event["type"] == "mouse-release":
        item = find_item_at_coordinate(event["pos"])
        state = State.get()
        # state.mouse_press_item = item
        if item is not None and item == state.mouse_press_item:
            item.activate(current_menu)
        state.mouse_press_item = None


def main_loop_iteration(
    input_func=default_input_func, render_func=default_render_func
):
    state = State.get()

    # if State.get().currently_ingame:
    #     print("currently ingame")
    #     #return False

    # print("main loop iteration")
    # time.sleep(0.1)
    # stop_loop = False

    # if State.get().idle_from and State.get().idle_from < get_current_time():
    #     #print(State.get().idle_from)
    #     if not Render.get().dirty:
    #         #print("waiting for events...")
    #         events = [pygame.event.wait()]
    #     else:
    #         events = pygame.event.get()
    # else:
    #     events = pygame.event.get()

    state.time = get_current_time()

    # Main.process()

    # t = State.get().time

    # if len(char_buffer) > 2 and t - char_buffer_last_updated > 0.5:
    # print(current_menu, isinstance(current_menu, ItemMenu))
    # if isinstance(current_menu, ItemMenu):
    #
    #    # if len(char_buffer) > 2 and t - char_buffer_last_updated > 0.5:
    #    if len(char_buffer) > 2:
    #        create_search_results_menu(char_buffer)
    #    if len(char_buffer) <= 2:
    #        # collapse search menu
    #
    #        if hasattr(current_menu, "search_text"):
    #            print("setting current_menu (something with search)")
    #            set_current_menu(current_menu.parent_menu)

    if state.hide_mouse_time and state.time > state.hide_mouse_time:
        if Mouse.focus:
            # keep cursor visible if mouse has focus
            pass
        else:
            state.hide_mouse_time = 0
            Mouse.set_visible(False)

    had_mouse_event = False
    for event in InputHandler.pop_all_text_events():
        if event["type"] == "text":
            # if key was handled as a virtual button, only allow this
            # character if already started typing something, or else
            # navigating with X-Arcade may start searching for games
            if InputHandler.peek_button():
                if len(current_menu.search_text) > 0:
                    character_press(event["text"])
                    # reset InputHandler so that virtual button
                    # is not processed, since we handled this press
                    # as a char
                    InputHandler.get_button()
            else:
                character_press(event["text"])
        elif event["type"] in ["mouse-motion", "mouse-press", "mouse-release"]:
            had_mouse_event = True
            handle_mouse_event(event)
        else:
            print("[WARNING] Unhandled event", event)

    AnimationSystem.update()
    NotificationRender.update()

    # if idle_events_only:
    #     # do not update State.get().idle_from
    #     pass
    # elif len(events) > 0:
    #     #print(events)
    #     if IDLE:
    #         State.get().idle_from = State.get().time + IDLE
    #         #print(State.get().idle_from)

    if Render.get().dirty or ALWAYS_RENDER:
        if Render.get().non_dirty_state:
            print("must start rendering again...")
        Render.get().non_dirty_state = False
        render_func()
    else:
        # if not Render.get().non_dirty_state:
        #     print("pause rendering!")
        # Render.get().non_dirty_state = True
        Render.get().dirty = True

    button = InputHandler.get_button()
    if button:
        print("InputHandler.get_button", button, InputHandler.last_device)
        GameCenter.register_user_activity()
        if input_func:
            input_func(button)
        Render.get().dirty = True
    if InputHandler.repeat_info:
        Render.get().dirty = True

    # if IDLE:
    if AnimationSystem.is_active():
        # State.get().idle_from = None
        Render.get().dirty = True
        # elif State.get().idle_from is None:
        #     State.get().idle_from = State.get().time + IDLE
    if not IDLE:
        Render.get().dirty = True
    # except KeyboardInterrupt:
    #     print "KeyboardInterrupt"
    #     return
    # return stop_loop
    if had_mouse_event:
        Render.get().dirty = True

    if not Render.get().display_sync:
        t = time.time()
        diff = t - Render.get().display_last_iteration
        # print(diff)
        frame_time = 1 / 60.0
        if diff < frame_time:
            # FIXME: use real refresh rate / frame time
            sleep_time = frame_time - diff
            time.sleep(sleep_time)
        Render.get().display_last_iteration = t

    return state.quit



# scan a set of file

from __future__ import print_function

import os
import fnmatch
import tempfile

from ScanFile import ScanFile

class FileScanner:

  def __init__(self, handler=None, ignore_filters=None, scanners=None,
               error_handler=None, ram_bytes=10 * 1024 * 1024,
               skip_handler=None, warning_handler=None):
    """the handler will be called with all the scanned files.
       the optional ignore_filters contains a list of glob pattern to
       ignore file names"""
    self.handler = handler
    self.error_handler = error_handler
    self.warning_handler = warning_handler
    self.skip_handler = skip_handler
    self.ignore_filters = ignore_filters
    self.scanners = scanners
    self.ram_bytes = ram_bytes

  def scan(self, path):
    """start scanning a path. either a file or directory"""
    if os.path.isdir(path):
      return self._scan_dir(path)
    elif os.path.isfile(path):
      return self._scan_file(path)
    else:
      return True

  def scan_obj(self, scan_file, check_ignore=True):
    """pass a ScanFile to check"""
    if check_ignore and self._is_ignored(scan_file.get_local_path()):
      return False
    # does a scanner match?
    sf = scan_file
    sc = self.scanners
    if sc is not None:
      for s in sc:
        if s.can_handle(sf):
          ok = s.handle(sf, self)
          sf.close()
          return ok
    # no match call user's handler
    ok = self._call_handler(sf)
    sf.close()
    return ok

  def _scan_dir(self, path):
    if self._is_ignored(path):
      return True
    for root, dirs, files in os.walk(path):
      for name in files:
        if not self._scan_file(os.path.join(root,name)):
          return False
      for name in dirs:
        if not self._scan_dir(os.path.join(root,name)):
          return False
    return True

  def _scan_file(self, path):
    if self._is_ignored(path):
      return True
    # build a scan file
    try:
      size = os.path.getsize(path)
      with open(path, "rb") as fobj:
        sf = ScanFile(path, fobj, size, True, True)
        return self.scan_obj(sf, False)
    except IOError as e:
      eh = self.error_handler
      if eh is not None:
        sf = ScanFile(path, None, 0)
        return eh(sf, e)
      else:
        # ignore error
        return True

  def _is_ignored(self, path):
    if self.ignore_filters is not None:
      base = os.path.basename(path)
      for f in self.ignore_filters:
        if fnmatch.fnmatch(base, f):
          return True
    return False

  def _call_handler(self, scan_file):
    if self.handler is not None:
      return self.handler(scan_file)
    else:
      return True

  def _call_skip_handler(self, scan_file):
    if self.skip_handler is not None:
      return self.skip_handler(scan_file)
    else:
      return True

  def promote_scan_file(self, scan_file, seekable=False, file_based=False):
    if not seekable and not file_base:
      return scan_file
    fb = file_based
    if not fb and seekable and scan_file.size > self.ram_bytes:
      fb = True
    sf = scan_file.create_clone(seekable, fb)
    scan_file.close()
    return sf

  def warn(self, scan_file, msg):
    wh = self.warning_handler
    if wh is not None:
      wh(scan_file, msg)


# mini test
if __name__ == '__main__':
  import sys
  ifs = ['*.txt']
  def handler(scan_file):
    print(scan_file)
    return True
  def error_handler(scan_file, error):
    print("FAILED:", scan_file, error)
    raise error
  fs = FileScanner(handler, ignore_filters=ifs, error_handler=error_handler)
  for a in sys.argv[1:]:
    fs.scan(a)



import fsbc.system
import fsui
from fsbc.util import unused

from .Constants import Constants
from .skin import LauncherTheme, Skin


class TabPanel(fsui.Panel):
    def __init__(self, parent, spacing=10):
        unused(spacing)
        fsui.Panel.__init__(self, parent, paintable=True)
        Skin.set_background_color(self)
        self.layout = fsui.HorizontalLayout()
        # self.layout.add_spacer(spacing)
        # self.layout.padding_left = 10
        # self.layout.padding_right = 10

        # self.set_background_color((0xdd, 0xdd, 0xdd))
        self.set_min_height(Constants.TAB_HEIGHT)

    def select_tab(self, index):
        counter = 0
        for child in self.layout.children:
            child = child.element
            if hasattr(child, "type"):
                if child.type == child.TYPE_TAB:
                    if counter == index:
                        child.select()
                    counter += 1

    def set_selected_tab(self, tab):
        for child in self.layout.children:
            child = child.element
            if hasattr(child, "type"):
                if child.type == child.TYPE_TAB:
                    if child == tab:
                        child.state = child.STATE_SELECTED
                        child.refresh()
                    elif child.state == child.STATE_SELECTED:
                        if child.group_id == tab.group_id:
                            child.state = child.STATE_NORMAL
                            child.refresh()

    def add(self, button, expand=False):
        self.layout.add(button, expand=expand)

    def add_spacer(self, spacer=0, expand=False):
        self.layout.add_spacer(spacer, 0, expand=expand)

    def on_paint(self):
        dc = self.create_dc()
        self.draw_background(self, dc)
        # self.draw_border(self, dc)

    @classmethod
    def draw_border(cls, widget, dc):
        size = widget.size()

        line_color_1 = Skin.get_background_color()
        if Skin.fws() or Skin.windows_10():
            from workspace.ui.theme import WorkspaceTheme

            theme = WorkspaceTheme.instance()
            line_color_1 = theme.title_separator_color
            line_color_2 = line_color_1
        elif line_color_1 is not None:
            line_color_1 = line_color_1.mix(fsui.Color(0xFF, 0xFF, 0xFF))
            line_color_2 = line_color_1
        else:
            if LauncherTheme.get().dark_mode:
                line_color_1 = fsui.Color(0xFF, 0xFF, 0xFF, 0x50)
            else:
                line_color_1 = fsui.Color(0xFF, 0xFF, 0xFF, 0xA0)
            line_color_2 = line_color_1

        # line_color_1 = fsui.Color(0xff, 0x00, 0x00, 0xff)
        # line_color_2 = fsui.Color(0x00, 0xff, 0x00, 0xff)

        dc.draw_line(0, size[1] - 2, size[0], size[1] - 2, line_color_1)
        dc.draw_line(0, size[1] - 1, size[0], size[1] - 1, line_color_2)

    @classmethod
    def draw_background(
        cls, widget, dc, selected=False, hover=False, button_style=True
    ):
        unused(button_style)
        if selected:
            cls.draw_selected_tab(widget, dc)
        else:
            cls.draw_border(widget, dc)

        size = widget.size()
        x = 0
        y = 0
        w = size[0]
        h = size[1] - 2

        if Skin.fws() or Skin.windows_10():
            from workspace.ui.theme import WorkspaceTheme

            theme = WorkspaceTheme.instance()
            white = fsui.Color(0xFF, 0xFF, 0xFF)
            if selected:
                bg_color = theme.window_background
                bd_color = theme.title_separator_color
            elif hover:
                bg_color = theme.window_background.copy().mix(
                    theme.title_background
                )
                bd_color = theme.title_separator_color.copy().mix(
                    theme.title_background
                )
            else:
                bg_color = theme.title_background
                bd_color = theme.title_background

            # dc.draw_rectangle(0, 0, w, h, bd_color)
            if selected or hover:
                dc.draw_vertical_gradient(0, 0, 2, size[1], white, bd_color)
                dc.draw_vertical_gradient(
                    size[0] - 2,
                    0,
                    2,
                    size[1],
                    theme.title_background,
                    bd_color,
                )
                # dc.draw_rectangle(2, 2, w - 4, h - 2, bg_color)
                dc.draw_vertical_gradient(
                    2,
                    0,
                    size[0] - 4,
                    size[1],
                    theme.title_background,
                    bg_color,
                )
                if hover and not selected:
                    dc.draw_rectangle(
                        0, size[1] - 2, size[0], 2, theme.title_separator_color
                    )
            else:
                dc.draw_rectangle(0, 0, w, h, bd_color)
            return

        if fsbc.system.macosx:
            # dc.draw_line(0, 0, w, 0, fsui.Color(198, 198, 198))
            # dc.draw_line(0, 0, w, 0, fsui.Color(188, 188, 188))
            dc.draw_line(0, 0, w, 0, fsui.Color(248, 248, 248))
            y += 1
            h -= 1

        if selected:
            x += 2
            w -= 4
            h += 2

        # if button_style and hover:
        #     x += 6
        #     y += 6
        #     w -= 12
        #     h -= 12

        color_1 = Skin.get_background_color()
        if fsbc.system.macosx and False:
            if selected:
                color_2 = color_1
                color_1 = fsui.Color(0xA7, 0xA7, 0xA7)
            elif hover:
                color_1 = fsui.Color(0xA7, 0xA7, 0xA7)
                color_2 = fsui.Color(0xEF, 0xEF, 0xEF)
            else:
                color_1 = fsui.Color(0xA7, 0xA7, 0xA7)
                color_2 = fsui.Color(0xC0, 0xC0, 0xC0)
        # elif Skin.adwaita():
        #     color_1 = fsui.Color(0xea, 0xe7, 0xe5)
        #     color_2 = fsui.Color(0xea, 0xe7, 0xe5)
        elif color_1 is not None:
            if selected:
                color_2 = color_1
            elif hover:
                color_2 = color_1.copy().lighten()
            else:
                color_2 = color_1.copy().darken(0.08)
        else:
            if selected:
                return
                # color_1 = fsui.Color(0x00, 0x00, 0x00, 0x00)
                # color_2 = color_1
            elif hover:
                color_1 = fsui.Color(0xFF, 0xFF, 0xFF, 0x00)
                color_2 = fsui.Color(0xFF, 0xFF, 0xFF, 0x40)
            else:
                color_1 = fsui.Color(0x00, 0x00, 0x00, 0x00)
                color_2 = fsui.Color(0x00, 0x00, 0x00, 0x20)
        dc.draw_vertical_gradient(x, y, w, h, color_1, color_2)

        if fsbc.system.macosx and False and not selected and not hover:
            dc.draw_line(
                x, y + h - 1, x + w, y + h - 1, fsui.Color(0xA8, 0xA8, 0xA8)
            )

    @classmethod
    def draw_selected_tab(cls, widget, dc):
        line_color_1 = Skin.get_background_color()
        if Skin.fws() or Skin.windows_10():
            from workspace.ui.theme import WorkspaceTheme

            theme = WorkspaceTheme.instance()
            line_color_1 = theme.title_separator_color
            line_color_2 = line_color_1
        elif fsbc.system.macosx and False:
            line_color_1 = fsui.Color(0xA7, 0xA7, 0xA7)
            line_color_2 = Skin.get_background_color().mix(
                fsui.Color(0xFF, 0xFF, 0xFF)
            )
        elif line_color_1 is not None:
            line_color_2 = Skin.get_background_color().mix(
                fsui.Color(0xFF, 0xFF, 0xFF)
            )
        else:
            if LauncherTheme.get().dark_mode:
                line_color_1 = fsui.Color(0xFF, 0xFF, 0xFF, 0x00)
                line_color_2 = fsui.Color(0xFF, 0xFF, 0xFF, 0x50)
            else:
                line_color_1 = fsui.Color(0xFF, 0xFF, 0xFF, 0x00)
                line_color_2 = fsui.Color(0xFF, 0xFF, 0xFF, 0xA0)

        size = widget.size()
        dc.draw_vertical_gradient(0, 0, 2, size[1], line_color_1, line_color_2)
        dc.draw_vertical_gradient(
            size[0] - 2, 0, 2, size[1], line_color_1, line_color_2
        )



from fsgs.drivers.messdriver import MessDriver


class MessSmsDriver(MessDriver):
    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "mastersystem",
    }

    PORTS = [
        {"description": "1st Controller", "types": [CONTROLLER]},
        {"description": "2nd Controller", "types": [CONTROLLER]},
    ]

    def mess_configure(self):
        self.mess_configure_cartridge()

    def mess_input_mapping(self, _):
        return {
            # "START": "P#_START",
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "1": "P#_BUTTON1",
            "2": "P#_BUTTON2",
        }

    def mess_offset_and_scale(self):
        if self.is_pal():
            return 0.0, 0.0, 1.082, 1.250
        return 0.0, 0.0, 1.082, 1.164

    def mess_romset(self):
        if self.is_pal():
            return "smspal", SMSPAL_ROMS
        else:
            return "sms", SMS_ROMS

    # def get_game_refresh_rate(self):
    #     # refresh rate retrieved from MESS
    #     return 59.922743


# noinspection SpellCheckingInspection
SMS_ROMS = {"3af7b66248d34eb26da40c92bf2fa4c73a46a051": "mpr-12808.ic2"}

# FIXME...
# noinspection SpellCheckingInspection
SMSPAL_ROMS = {
    "3af7b66248d34eb26da40c92bf2fa4c73a46a051": "mpr-12808.ic2",
    "6aca0e3dffe461ba1cb11a86cd4caf5b97e1b8df": "sonbios.rom",
}



from OpenGL import GL


def fs_emu_blending(enable):
    if enable:
        GL.glEnable(GL.GL_BLEND)
    else:
        GL.glDisable(GL.GL_BLEND)


def fs_emu_texturing(enable):
    if enable:
        GL.glEnable(GL.GL_TEXTURE_2D)
    else:
        GL.glDisable(GL.GL_TEXTURE_2D)


# class GL:
#     # These constants are repeated here to aid code inspection tools (in
#     # PyOpenGL they are expanded (run-time) from a string).

#     GL_BLEND = 0xBE2
#     GL_TEXTURE_2D = 0xDE1
#     GL_COMPILE_STATUS = 0x8B81
#     GL_LINK_STATUS = 0x8B82
#     GL_FRAGMENT_SHADER = 0x8B30
#     GL_VERTEX_SHADER = 0x8B31
#     GL_ONE = 0x1
#     GL_ONE_MINUS_SRC_COLOR = 0x301
#     GL_SRC_ALPHA = 0x302
#     GL_ONE_MINUS_SRC_ALPHA = 0x303
#     GL_RGB = 0x1907
#     GL_RGBA = 0x1908
#     GL_BGRA = 0x80E1
#     GL_TEXTURE_MAG_FILTER = 0x2800
#     GL_TEXTURE_MIN_FILTER = 0x2801
#     GL_FRAMEBUFFER_EXT = 0x8D40
#     GL_COLOR_ATTACHMENT0_EXT = 0x8CE0
#     GL_RENDERBUFFER_EXT = 0x8D41
#     GL_DEPTH_ATTACHMENT_EXT = 0x8D00
#     GL_LINEAR = 0x2601
#     GL_FRAMEBUFFER_COMPLETE_EXT = 0x8CD5
#     GL_DEPTH_COMPONENT24 = 0x81A6
#     GL_QUADS = 0x7
#     GL_DEPTH_TEST = 0xB71
#     GL_TEXTURE_2D_ARB = 0x207A
#     GL_COLOR_BUFFER_BIT = 0x4000
#     GL_DEPTH_BUFFER_BIT = 0x100
#     GL_MODELVIEW = 0x1700
#     GL_PROJECTION = 0x1701
#     GL_LIGHT_MODEL_TWO_SIDE = 0xB52
#     GL_LIGHT_MODEL_LOCAL_VIEWER = 0xB51
#     GL_LIGHT_MODEL_AMBIENT = 0xB53
#     GL_SEPARATE_SPECULAR_COLOR = 0x81FA
#     GL_LIGHT_MODEL_COLOR_CONTROL = 0x81F8
#     GL_LIGHTING = 0xB50
#     GL_LIGHT0 = 0x4000
#     GL_LIGHT1 = 0x4001
#     GL_LIGHT2 = 0x4002
#     GL_LIGHT3 = 0x4003
#     GL_LIGHT4 = 0x4004
#     GL_LIGHT5 = 0x4005
#     GL_LIGHT6 = 0x4006
#     GL_LIGHT7 = 0x4007
#     GL_AMBIENT = 0x1200
#     GL_DIFFUSE = 0x1201
#     GL_SPECULAR = 0x1202
#     GL_SHININESS = 0x1601
#     GL_EMISSION = 0x1600
#     GL_POSITION = 0x1203
#     GL_FRONT = 0x404
#     GL_BACK = 0x405
#     GL_POLYGON_OFFSET_FILL = 0x8037
#     GL_CLAMP_TO_EDGE = 0x812F

#     def __init__(self):
#         self._loaded = False
#         self._first_time = True

#     def load(self):
#         if self._loaded:
#             return
#         if self._first_time:
#             import locale

#             # we must set the LC_NUMERIC locale to C, because PyOpenGL has
#             # a problem if comma is used as a decimal separator (problem
#             # parsing the OpenGL version)
#             locale.setlocale(locale.LC_NUMERIC, str("C"))
#         from .pyopengl import filter_global, pyopengl_globals

#         for name in list(pyopengl_globals()):
#             if filter_global(name):
#                 setattr(gl, name, pyopengl_globals()[name])
#         if self._first_time:
#             self.log_opengl_info()
#         self._loaded = True
#         self._first_time = False

#     def unload(self):
#         for name in list(self.__dict__.keys()):
#             if name.lower().startswith("gl"):
#                 delattr(self, name)
#         self._loaded = False

#     def log_opengl_info(self):
#         print("[OPENGL] PyOpenGL information:")
#         print("[OPENGL]", sys.modules["OpenGL"])
#         print("[OPENGL] Vendor:", self.glGetString(self.GL_VENDOR))
#         print("[OPENGL] Renderer:", self.glGetString(self.GL_RENDERER))
#         print("[OPENGL] Version:", self.glGetString(self.GL_VERSION))


# gl = GL()

# for __x in list(globals()):
#     if filter_global(__x):
#         setattr(gl, __x, globals()[__x])


# if False:
#     # this fixes ImportError: No module named win32 when used with py2exe
#     # and other problems related to required modules missing from pyOpenGL
#     # noinspection PyUnresolvedReferences
#     import ctypes

#     # noinspection PyUnresolvedReferences
#     import logging

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays._buffers

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays._numeric

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays._strings

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.arraydatatype

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.arrayhelpers

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.buffers

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.ctypesarrays

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.ctypesparameters

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.ctypespointers

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.formathandler

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.lists

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.nones

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.numbers

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.numeric

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.numericnames

#     # noinspection PyUnresolvedReferences
#     # import OpenGL.arrays.numpymodule
#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.strings

#     # noinspection PyUnresolvedReferences
#     import OpenGL.arrays.vbo

#     # noinspection PyUnresolvedReferences
#     import OpenGL.platform.win32



# Automatically generated - do not edit by hand


# noinspection SpellCheckingInspection
class Option(object):
    """Constants for option names."""

    CONFIG_REFRESH = "__config_refresh"
    NETPLAY_GAME = "__netplay_game"
    A2600_DATABASE = "a2600_database"
    A5200_DATABASE = "a5200_database"
    A7800_A78_HEADER = "a7800_a78_header"
    A7800_DATABASE = "a7800_database"
    A7800_MODEL = "a7800_model"
    A7800_PORT_1_TYPE = "a7800_port_1_type"
    A7800_PORT_2_TYPE = "a7800_port_2_type"
    ACCELERATOR = "accelerator"
    ACCELERATOR_MEMORY = "accelerator_memory"
    ACCELERATOR_ROM = "accelerator_rom"
    ACCURACY = "accuracy"
    ALIGN_X = "align_x"
    ALIGN_Y = "align_y"
    ALWAYS_ON_TOP = "always_on_top"
    AMIGA_DATABASE = "amiga_database"
    AMIGA_DRIVER = "amiga_driver"
    AMIGA_MODEL = "amiga_model"
    ARCADE_DATABASE = "arcade_database"
    ARCADE_FULLSCREEN = "arcade_fullscreen"
    ARCADE_INITIAL_FAVORITES = "arcade_initial_favorites"
    ARCADE_MAXIMIZED = "arcade_maximized"
    ARCADE_SEARCH = "arcade_search"
    ARCADE_SHUTDOWN = "arcade_shutdown"
    ARCADE_THEME = "arcade_theme"
    ASSUME_REFRESH_RATE = "assume_refresh_rate"
    ATARI_DATABASE = "atari_database"
    ATARI_MODEL = "atari_model"
    AUDIO_BUFFER_TARGET_BYTES = "audio_buffer_target_bytes"
    AUDIO_BUFFER_TARGET_SIZE = "audio_buffer_target_size"
    AUDIO_DRIVER = "audio_driver"
    AUDIO_FREQUENCY = "audio_frequency"
    AUTO_LOAD = "auto_load"
    AUTO_QUIT = "auto_quit"
    AUTOMATIC_INPUT_GRAB = "automatic_input_grab"
    BASE_DIR = "base_dir"
    BEZEL = "bezel"
    BLIZZARD_SCSI_KIT = "blizzard_scsi_kit"
    BORDER = "border"
    BSDSOCKET_LIBRARY = "bsdsocket_library"
    BUILTIN_CONFIGS = "builtin_configs"
    C64_DATABASE = "c64_database"
    C64_MODEL = "c64_model"
    C64_PALETTE = "c64_palette"
    C64_PORT_1_TYPE = "c64_port_1_type"
    C64_PORT_2_TYPE = "c64_port_2_type"
    CARTRIDGE_SLOT = "cartridge_slot"
    CD32_DATABASE = "cd32_database"
    CDFS = "cdfs"
    CDROM_DRIVE_0 = "cdrom_drive_0"
    CDROM_DRIVE_0_DELAY = "cdrom_drive_0_delay"
    CDROM_DRIVE_COUNT = "cdrom_drive_count"
    CDROM_IMAGE_0 = "cdrom_image_0"
    CDROM_IMAGE_1 = "cdrom_image_1"
    CDROM_IMAGE_10 = "cdrom_image_10"
    CDROM_IMAGE_11 = "cdrom_image_11"
    CDROM_IMAGE_12 = "cdrom_image_12"
    CDROM_IMAGE_13 = "cdrom_image_13"
    CDROM_IMAGE_14 = "cdrom_image_14"
    CDROM_IMAGE_15 = "cdrom_image_15"
    CDROM_IMAGE_16 = "cdrom_image_16"
    CDROM_IMAGE_17 = "cdrom_image_17"
    CDROM_IMAGE_18 = "cdrom_image_18"
    CDROM_IMAGE_19 = "cdrom_image_19"
    CDROM_IMAGE_2 = "cdrom_image_2"
    CDROM_IMAGE_3 = "cdrom_image_3"
    CDROM_IMAGE_4 = "cdrom_image_4"
    CDROM_IMAGE_5 = "cdrom_image_5"
    CDROM_IMAGE_6 = "cdrom_image_6"
    CDROM_IMAGE_7 = "cdrom_image_7"
    CDROM_IMAGE_8 = "cdrom_image_8"
    CDROM_IMAGE_9 = "cdrom_image_9"
    CDROMS_DIR = "cdroms_dir"
    CDTV_DATABASE = "cdtv_database"
    CHEATS = "cheats"
    CHIP_MEMORY = "chip_memory"
    CLIPBOARD_SHARING = "clipboard_sharing"
    COMMAND = "command"
    CONFIG_NAME = "config_name"
    CONSOLE_DEBUGGER = "console_debugger"
    CONTROLLERS_DIR = "controllers_dir"
    CPC_DATABASE = "cpc_database"
    CPU = "cpu"
    CPU_IDLE = "cpu_idle"
    CPUBOARD_FLASH_EXT_FILE = "cpuboard_flash_ext_file"
    CPUBOARD_FLASH_FILE = "cpuboard_flash_file"
    CROP = "crop"
    CUE_SHEETS = "cue_sheets"
    CURSOR = "cursor"
    DATABASE_AUTH = "database_auth"
    DATABASE_EMAIL = "database_email"
    DATABASE_FEATURE = "database_feature"
    DATABASE_LOCKER = "database_locker"
    DATABASE_PASSWORD = "database_password"
    DATABASE_SERVER = "database_server"
    DATABASE_SHOW_ADULT = "database_show_adult"
    DATABASE_SHOW_GAMES = "database_show_games"
    DATABASE_SHOW_UNPUBLISHED = "database_show_unpublished"
    DATABASE_USERNAME = "database_username"
    DETERMINISTIC = "deterministic"
    DEVELOPER_MODE = "developer_mode"
    DEVICE_ID = "device_id"
    DISABLE_ASPECT_CORRECTION = "disable_aspect_correction"
    DONGLE_TYPE = "dongle_type"
    DOS_DATABASE = "dos_database"
    DOS_EMULATOR = "dos_emulator"
    DOSBOX_CPU_CORE = "dosbox_cpu_core"
    DOSBOX_CPU_CPUTYPE = "dosbox_cpu_cputype"
    DOSBOX_CPU_CYCLES = "dosbox_cpu_cycles"
    DOSBOX_GUS_GUS = "dosbox_gus_gus"
    DOSBOX_MACHINE = "dosbox_machine"
    DOSBOX_MEMSIZE = "dosbox_memsize"
    DOSBOX_SBLASTER_IRQ = "dosbox_sblaster_irq"
    DOSBOX_SBLASTER_OPLRATE = "dosbox_sblaster_oplrate"
    DOSBOX_SBLASTER_SBBASE = "dosbox_sblaster_sbbase"
    DOSBOX_SBLASTER_SBTYPE = "dosbox_sblaster_sbtype"
    EFFECT = "effect"
    EMULATOR_PID_FILE = "emulator_pid_file"
    END_CONFIG = "end_config"
    EXPECT_VERSION = "expect_version"
    FADE_IN_DURATION = "fade_in_duration"
    FADE_OUT_DURATION = "fade_out_duration"
    FAKE_JOYSTICKS = "fake_joysticks"
    FAKE_VERSION = "fake_version"
    FAST_MEMORY = "fast_memory"
    FILE_LIST = "file_list"
    FLASH_MEMORY_DIR = "flash_memory_dir"
    FLOPPIES_DIR = "floppies_dir"
    FLOPPY_DRIVE_0 = "floppy_drive_0"
    FLOPPY_DRIVE_0_SOUNDS = "floppy_drive_0_sounds"
    FLOPPY_DRIVE_1 = "floppy_drive_1"
    FLOPPY_DRIVE_1_SOUNDS = "floppy_drive_1_sounds"
    FLOPPY_DRIVE_2 = "floppy_drive_2"
    FLOPPY_DRIVE_2_SOUNDS = "floppy_drive_2_sounds"
    FLOPPY_DRIVE_3 = "floppy_drive_3"
    FLOPPY_DRIVE_3_SOUNDS = "floppy_drive_3_sounds"
    FLOPPY_DRIVE_COUNT = "floppy_drive_count"
    FLOPPY_DRIVE_SPEED = "floppy_drive_speed"
    FLOPPY_DRIVE_VOLUME = "floppy_drive_volume"
    FLOPPY_DRIVE_VOLUME_EMPTY = "floppy_drive_volume_empty"
    FLOPPY_IMAGE_0 = "floppy_image_0"
    FLOPPY_IMAGE_1 = "floppy_image_1"
    FLOPPY_IMAGE_10 = "floppy_image_10"
    FLOPPY_IMAGE_11 = "floppy_image_11"
    FLOPPY_IMAGE_12 = "floppy_image_12"
    FLOPPY_IMAGE_13 = "floppy_image_13"
    FLOPPY_IMAGE_14 = "floppy_image_14"
    FLOPPY_IMAGE_15 = "floppy_image_15"
    FLOPPY_IMAGE_16 = "floppy_image_16"
    FLOPPY_IMAGE_17 = "floppy_image_17"
    FLOPPY_IMAGE_18 = "floppy_image_18"
    FLOPPY_IMAGE_19 = "floppy_image_19"
    FLOPPY_IMAGE_2 = "floppy_image_2"
    FLOPPY_IMAGE_3 = "floppy_image_3"
    FLOPPY_IMAGE_4 = "floppy_image_4"
    FLOPPY_IMAGE_5 = "floppy_image_5"
    FLOPPY_IMAGE_6 = "floppy_image_6"
    FLOPPY_IMAGE_7 = "floppy_image_7"
    FLOPPY_IMAGE_8 = "floppy_image_8"
    FLOPPY_IMAGE_9 = "floppy_image_9"
    FLOPPY_OVERLAYS_DIR = "floppy_overlays_dir"
    FMV_ROM = "fmv_rom"
    FORCE_ASPECT = "force_aspect"
    FPU = "fpu"
    FRAME_TIME = "frame_time"
    FREEZER_CARTRIDGE = "freezer_cartridge"
    FSAA = "fsaa"
    FULL_KEYBOARD = "full_keyboard"
    FULLSCREEN = "fullscreen"
    FULLSCREEN_HEIGHT = "fullscreen_height"
    FULLSCREEN_MODE = "fullscreen_mode"
    FULLSCREEN_WIDTH = "fullscreen_width"
    G_SYNC = "g_sync"
    GAME_LIST_UUID = "game_list_uuid"
    GAME_MODE = "game_mode"
    GAME_NAME = "game_name"
    GB_DATABASE = "gb_database"
    GB_PORT_1_TYPE = "gb_port_1_type"
    GBA_DATABASE = "gba_database"
    GBA_GAMMA = "gba_gamma"
    GBA_PORT_1_TYPE = "gba_port_1_type"
    GBC_DATABASE = "gbc_database"
    GBC_PORT_1_TYPE = "gbc_port_1_type"
    GOVERNOR_WARNING = "governor_warning"
    GRAB_INPUT = "grab_input"
    GRAPHICS_CARD = "graphics_card"
    GRAPHICS_CARD_ROM = "graphics_card_rom"
    GRAPHICS_MEMORY = "graphics_memory"
    HARD_DRIVE_0 = "hard_drive_0"
    HARD_DRIVE_0_CONTROLLER = "hard_drive_0_controller"
    HARD_DRIVE_0_FILE_SYSTEM = "hard_drive_0_file_system"
    HARD_DRIVE_0_LABEL = "hard_drive_0_label"
    HARD_DRIVE_0_PRIORITY = "hard_drive_0_priority"
    HARD_DRIVE_0_READ_ONLY = "hard_drive_0_read_only"
    HARD_DRIVE_0_TYPE = "hard_drive_0_type"
    HARD_DRIVE_1 = "hard_drive_1"
    HARD_DRIVE_2 = "hard_drive_2"
    HARD_DRIVE_3 = "hard_drive_3"
    HARD_DRIVE_4 = "hard_drive_4"
    HARD_DRIVE_5 = "hard_drive_5"
    HARD_DRIVE_6 = "hard_drive_6"
    HARD_DRIVE_7 = "hard_drive_7"
    HARD_DRIVE_8 = "hard_drive_8"
    HARD_DRIVE_9 = "hard_drive_9"
    HARD_DRIVES_DIR = "hard_drives_dir"
    INITIAL_INPUT_GRAB = "initial_input_grab"
    INPUT_GRAB = "input_grab"
    INT_ACCELERATOR_NAME = "int_accelerator_name"
    INT_BOGOMEM_SIZE = "int_bogomem_size"
    INT_CHIPMEM_SIZE = "int_chipmem_size"
    INT_CHIPSET_NAME = "int_chipset_name"
    INT_CPU_NAME = "int_cpu_name"
    INT_CPUBOARDMEM1_SIZE = "int_cpuboardmem1_size"
    INT_DEFAULT_FLOPPY_TYPE = "int_default_floppy_type"
    INT_FASTMEM_SIZE = "int_fastmem_size"
    INT_GRAPHICS_CARD_BUS = "int_graphics_card_bus"
    INT_GRAPHICS_CARD_NAME = "int_graphics_card_name"
    INT_KICKSTART_EXT_SHA1 = "int_kickstart_ext_sha1"
    INT_KICKSTART_REVISION = "int_kickstart_revision"
    INT_KICKSTART_SHA1 = "int_kickstart_sha1"
    INT_KICKSTART_VERSION = "int_kickstart_version"
    INT_MBRESMEM_LOW_SIZE = "int_mbresmem_low_size"
    INT_MODEL = "int_model"
    INT_MODEL_NAME = "int_model_name"
    INT_PPC_MODEL = "int_ppc_model"
    INT_UAE_BOOT_ROM = "int_uae_boot_rom"
    INT_Z3CHIPMEM_SIZE = "int_z3chipmem_size"
    INT_Z3FASTMEM_SIZE = "int_z3fastmem_size"
    IRC_NICK = "irc_nick"
    IRC_SERVER = "irc_server"
    JIT_COMPILER = "jit_compiler"
    JIT_MEMORY = "jit_memory"
    JOYSTICK_PORT_0 = "joystick_port_0"
    JOYSTICK_PORT_0_AUTOFIRE = "joystick_port_0_autofire"
    JOYSTICK_PORT_0_AUTOSWITCH = "joystick_port_0_autoswitch"
    JOYSTICK_PORT_0_MODE = "joystick_port_0_mode"
    JOYSTICK_PORT_1 = "joystick_port_1"
    JOYSTICK_PORT_1_AUTOFIRE = "joystick_port_1_autofire"
    JOYSTICK_PORT_1_MODE = "joystick_port_1_mode"
    JOYSTICK_PORT_2 = "joystick_port_2"
    JOYSTICK_PORT_2_AUTOFIRE = "joystick_port_2_autofire"
    JOYSTICK_PORT_2_MODE = "joystick_port_2_mode"
    JOYSTICK_PORT_3 = "joystick_port_3"
    JOYSTICK_PORT_3_AUTOFIRE = "joystick_port_3_autofire"
    JOYSTICK_PORT_3_MODE = "joystick_port_3_mode"
    KEEP_ASPECT = "keep_aspect"
    KEYBOARD_INPUT_GRAB = "keyboard_input_grab"
    KEYBOARD_KEY_BACKSLASH = "keyboard_key_backslash"
    KEYBOARD_KEY_EQUALS = "keyboard_key_equals"
    KEYBOARD_KEY_INSERT = "keyboard_key_insert"
    KEYBOARD_KEY_LESS = "keyboard_key_less"
    KICKSTART_EXT_FILE = "kickstart_ext_file"
    KICKSTART_FILE = "kickstart_file"
    KICKSTART_SETUP = "kickstart_setup"
    KICKSTARTS_DIR = "kickstarts_dir"
    LANGUAGE = "language"
    LAUNCHER_CLOSE_BUTTONS = "launcher_close_buttons"
    LAUNCHER_CONFIG_FEATURE = "launcher_config_feature"
    LAUNCHER_FONT_SIZE = "launcher_font_size"
    LAUNCHER_LOG_CONFIG = "launcher_log_config"
    LAUNCHER_SETUP_WIZARD_FEATURE = "launcher_setup_wizard_feature"
    LAUNCHER_THEME = "launcher_theme"
    LINE_DOUBLING = "line_doubling"
    LOAD_STATE = "load_state"
    LOCALIZATION = "localization"
    LOG_AUTOSCALE = "log_autoscale"
    LOG_BSDSOCKET = "log_bsdsocket"
    LOG_FLUSH = "log_flush"
    LOG_INPUT = "log_input"
    LOG_QUERY_PLANS = "log_query_plans"
    LOGS_DIR = "logs_dir"
    LOW_LATENCY_VSYNC = "low_latency_vsync"
    LOW_RESOLUTION = "low_resolution"
    MAME_ARTWORK = "mame_artwork"
    MEDNAFEN_AUDIO_BUFFER = "mednafen_audio_buffer"
    MEDNAFEN_AUDIO_DRIVER = "mednafen_audio_driver"
    MEDNAFEN_DEINTERLACER = "mednafen_deinterlacer"
    MEDNAFEN_TEMPORAL_BLUR = "mednafen_temporal_blur"
    MIDDLE_CLICK_UNGRAB = "middle_click_ungrab"
    MIN_FIRST_LINE_NTSC = "min_first_line_ntsc"
    MIN_FIRST_LINE_PAL = "min_first_line_pal"
    MODEL = "model"
    MONITOR = "monitor"
    MOTHERBOARD_RAM = "motherboard_ram"
    MOUSE_INTEGRATION = "mouse_integration"
    MOUSE_SPEED = "mouse_speed"
    MSX_DATABASE = "msx_database"
    N64_DATABASE = "n64_database"
    NDS_DATABASE = "nds_database"
    NEOGEO_DATABASE = "neogeo_database"
    NEOGEO_MODEL = "neogeo_model"
    NEOGEO_PORT_1_TYPE = "neogeo_port_1_type"
    NEOGEO_PORT_2_TYPE = "neogeo_port_2_type"
    NES_DATABASE = "nes_database"
    NES_DRIVER = "nes_driver"
    NES_EMULATOR = "nes_emulator"
    NES_INES_HEADER = "nes_ines_header"
    NES_MODEL = "nes_model"
    NES_PORT_1_TYPE = "nes_port_1_type"
    NES_PORT_2_TYPE = "nes_port_2_type"
    NES_PORT_3_TYPE = "nes_port_3_type"
    NES_PORT_4_TYPE = "nes_port_4_type"
    NETPLAY_FEATURE = "netplay_feature"
    NETPLAY_PASSWORD = "netplay_password"
    NETPLAY_PORT = "netplay_port"
    NETPLAY_SERVER = "netplay_server"
    NETPLAY_TAG = "netplay_tag"
    NETWORK_CARD = "network_card"
    NGC_DATABASE = "ngc_database"
    NOTIFICATION_DURATION = "notification_duration"
    NTSC_MODE = "ntsc_mode"
    OFFLINE_MODE = "offline_mode"
    ORIENTATION = "orientation"
    PLATFORM = "platform"
    PLATFORMS_FEATURE = "platforms_feature"
    PRIMARY_JOYSTICK = "primary_joystick"
    PSX_DATABASE = "psx_database"
    PSX_MODEL = "psx_model"
    PSX_PORT_1_TYPE = "psx_port_1_type"
    PSX_PORT_2_TYPE = "psx_port_2_type"
    PSX_PRELOAD = "psx_preload"
    QUICK_SETTINGS = "quick_settings"
    QUICK_SETTINGS_OPTIONS = "quick_settings_options"
    RAW_INPUT = "raw_input"
    REFRESH_RATE = "refresh_rate"
    RELATIVE_PATHS = "relative_paths"
    RELATIVE_TEMP_FEATURE = "relative_temp_feature"
    RETROARCH_AUDIO_BUFFER = "retroarch_audio_buffer"
    RTG_SCANLINES = "rtg_scanlines"
    RTG_VIEWPORT = "rtg_viewport"
    SAVE_DISK = "save_disk"
    SAVE_STATE_COMPRESSION = "save_state_compression"
    SAVE_STATES = "save_states"
    SAVE_STATES_DIR = "save_states_dir"
    SBI_DATA = "sbi_data"
    SBLASTER_IRQ = "sblaster_irq"
    SCALE = "scale"
    SCALE_X = "scale_x"
    SCALE_Y = "scale_y"
    SCANLINES = "scanlines"
    SCANLINES_DARK = "scanlines_dark"
    SCANLINES_LIGHT = "scanlines_light"
    SCREENSHOTS_OUTPUT_DIR = "screenshots_output_dir"
    SCREENSHOTS_OUTPUT_MASK = "screenshots_output_mask"
    SCREENSHOTS_OUTPUT_PREFIX = "screenshots_output_prefix"
    SEARCH_PATH = "search_path"
    SECONDARY_JOYSTICK = "secondary_joystick"
    SERIAL_PORT = "serial_port"
    SGG_DATABASE = "sgg_database"
    SGG_PORT_1_TYPE = "sgg_port_1_type"
    SHADER = "shader"
    SLOW_MEMORY = "slow_memory"
    SMD_DATABASE = "smd_database"
    SMD_EMULATOR = "smd_emulator"
    SMD_MODEL = "smd_model"
    SMD_PORT_1_TYPE = "smd_port_1_type"
    SMD_PORT_2_TYPE = "smd_port_2_type"
    SMOOTHING = "smoothing"
    SMS_DATABASE = "sms_database"
    SMS_DRIVER = "sms_driver"
    SNAPSHOT_FILE = "snapshot_file"
    SNES_DATABASE = "snes_database"
    SNES_PORT_1_TYPE = "snes_port_1_type"
    SNES_PORT_2_TYPE = "snes_port_2_type"
    SOUND_CARD = "sound_card"
    SRAM_TYPE = "sram_type"
    STATE_DIR = "state_dir"
    STATE_DIR_NAME = "state_dir_name"
    STEREO_SEPARATION = "stereo_separation"
    STRETCH = "stretch"
    SUB_TITLE = "sub_title"
    SWAP_CTRL_KEYS = "swap_ctrl_keys"
    SYNC = "sync"
    TAPE_DRIVE_0 = "tape_drive_0"
    TEXTURE_FILTER = "texture_filter"
    TEXTURE_FORMAT = "texture_format"
    TG16_DATABASE = "tg16_database"
    TGCD_DATABASE = "tgcd_database"
    THEME = "theme"
    THEME_CUSTOM_0 = "theme_custom_0"
    THEME_CUSTOM_0_X = "theme_custom_0_x"
    THEME_CUSTOM_0_Y = "theme_custom_0_y"
    THEME_FADE_COLOR = "theme_fade_color"
    THEME_FLOOR_COLOR_1 = "theme_floor_color_1"
    THEME_FLOOR_COLOR_2 = "theme_floor_color_2"
    THEME_HEADING_COLOR = "theme_heading_color"
    THEME_OVERLAY_IMAGE = "theme_overlay_image"
    THEME_WALL_COLOR_1 = "theme_wall_color_1"
    THEME_WALL_COLOR_2 = "theme_wall_color_2"
    TITLE = "title"
    TURBO_LOAD = "turbo_load"
    UAE_A1000RAM = "uae_a1000ram"
    UAE_A2065 = "uae_a2065"
    UAE_A3000MEM_SIZE = "uae_a3000mem_size"
    UAE_AGNUS_BLTBUSYBUG = "uae_agnus_bltbusybug"
    UAE_BOGOMEM_FAST = "uae_bogomem_fast"
    UAE_BOGOMEM_SIZE = "uae_bogomem_size"
    UAE_BSDSOCKET_EMU = "uae_bsdsocket_emu"
    UAE_CART = "uae_cart"
    UAE_CART_FILE = "uae_cart_file"
    UAE_CART_INTERNAL = "uae_cart_internal"
    UAE_CD32C2P = "uae_cd32c2p"
    UAE_CD32CD = "uae_cd32cd"
    UAE_CD32FMV = "uae_cd32fmv"
    UAE_CD32NVRAM = "uae_cd32nvram"
    UAE_CDTVCD = "uae_cdtvcd"
    UAE_CDTVRAM = "uae_cdtvram"
    UAE_CDTVRAMCARD = "uae_cdtvramcard"
    UAE_CHIPMEM_SIZE = "uae_chipmem_size"
    UAE_CHIPSET = "uae_chipset"
    UAE_CHIPSET_COMPATIBLE = "uae_chipset_compatible"
    UAE_CIAATOD = "uae_ciaatod"
    UAE_COLLISION_LEVEL = "uae_collision_level"
    UAE_CPU_24BIT_ADDRESSING = "uae_cpu_24bit_addressing"
    UAE_CPU_COMPATIBLE = "uae_cpu_compatible"
    UAE_CPU_FREQUENCY = "uae_cpu_frequency"
    UAE_CPU_MODEL = "uae_cpu_model"
    UAE_CPU_MULTIPLIER = "uae_cpu_multiplier"
    UAE_CPU_RESET_PAUSE = "uae_cpu_reset_pause"
    UAE_CPU_SPEED = "uae_cpu_speed"
    UAE_CPU_THROTTLE = "uae_cpu_throttle"
    UAE_CPU_TYPE = "uae_cpu_type"
    UAE_CPUBOARD_TYPE = "uae_cpuboard_type"
    UAE_CPUBOARDMEM1_SIZE = "uae_cpuboardmem1_size"
    UAE_DENISE_NOEHB = "uae_denise_noehb"
    UAE_DONGLE = "uae_dongle"
    UAE_FASTMEM2_SIZE = "uae_fastmem2_size"
    UAE_FASTMEM2_SIZE_K = "uae_fastmem2_size_k"
    UAE_FASTMEM_AUTOCONFIG = "uae_fastmem_autoconfig"
    UAE_FASTMEM_SIZE = "uae_fastmem_size"
    UAE_FASTMEM_SIZE_K = "uae_fastmem_size_k"
    UAE_FATGARY = "uae_fatgary"
    UAE_FILESYS_INJECT_ICONS = "uae_filesys_inject_icons"
    UAE_FILESYS_INJECT_ICONS_DRAWER = "uae_filesys_inject_icons_drawer"
    UAE_FILESYS_INJECT_ICONS_PROJECT = "uae_filesys_inject_icons_project"
    UAE_FILESYS_INJECT_ICONS_TOOL = "uae_filesys_inject_icons_tool"
    UAE_FINEGRAIN_CPU_SPEED = "uae_finegrain_cpu_speed"
    UAE_FLOPPY0 = "uae_floppy0"
    UAE_FLOPPY0TYPE = "uae_floppy0type"
    UAE_FLOPPY1 = "uae_floppy1"
    UAE_FLOPPY1TYPE = "uae_floppy1type"
    UAE_FLOPPY2 = "uae_floppy2"
    UAE_FLOPPY2TYPE = "uae_floppy2type"
    UAE_FLOPPY3 = "uae_floppy3"
    UAE_FLOPPY3TYPE = "uae_floppy3type"
    UAE_FORCE_0X10000000_Z3 = "uae_force_0x10000000_z3"
    UAE_FPU_MODEL = "uae_fpu_model"
    UAE_GFX_LINEMODE = "uae_gfx_linemode"
    UAE_GFXCARD_HARDWARE_SPRITE = "uae_gfxcard_hardware_sprite"
    UAE_GFXCARD_HARDWARE_VBLANK = "uae_gfxcard_hardware_vblank"
    UAE_GFXCARD_SIZE = "uae_gfxcard_size"
    UAE_GFXCARD_TYPE = "uae_gfxcard_type"
    UAE_ICS_AGNUS = "uae_ics_agnus"
    UAE_IDE = "uae_ide"
    UAE_IMMEDIATE_BLITS = "uae_immediate_blits"
    UAE_JOYPORT0 = "uae_joyport0"
    UAE_JOYPORT0AUTOFIRE = "uae_joyport0autofire"
    UAE_JOYPORT0MODE = "uae_joyport0mode"
    UAE_JOYPORT1 = "uae_joyport1"
    UAE_JOYPORT1AUTOFIRE = "uae_joyport1autofire"
    UAE_JOYPORT1MODE = "uae_joyport1mode"
    UAE_KSMIRROR_A8 = "uae_ksmirror_a8"
    UAE_KSMIRROR_E0 = "uae_ksmirror_e0"
    UAE_MAPROM = "uae_maprom"
    UAE_MBRESMEM_SIZE = "uae_mbresmem_size"
    UAE_MEGACHIPMEM_SIZE = "uae_megachipmem_size"
    UAE_MMU_MODEL = "uae_mmu_model"
    UAE_NATIVE_CODE = "uae_native_code"
    UAE_PCMCIA = "uae_pcmcia"
    UAE_PPC_CPU_IDLE = "uae_ppc_cpu_idle"
    UAE_PPC_MODEL = "uae_ppc_model"
    UAE_RAMSEY = "uae_ramsey"
    UAE_RESETWARNING = "uae_resetwarning"
    UAE_RTC = "uae_rtc"
    UAE_SANA2 = "uae_sana2"
    UAE_SCSI_A2091 = "uae_scsi_a2091"
    UAE_SCSI_A3000 = "uae_scsi_a3000"
    UAE_SCSI_A4000T = "uae_scsi_a4000t"
    UAE_SCSI_A4091 = "uae_scsi_a4091"
    UAE_SCSI_CDTV = "uae_scsi_cdtv"
    UAE_SCSIDEV_MODE = "uae_scsidev_mode"
    UAE_SLIRP_IMPLEMENTATION = "uae_slirp_implementation"
    UAE_SLIRP_PORTS = "uae_slirp_ports"
    UAE_SLIRP_REDIR = "uae_slirp_redir"
    UAE_SOUND_INTERPOL = "uae_sound_interpol"
    UAE_SOUND_OUTPUT = "uae_sound_output"
    UAE_SOUND_STEREO_SEPARATION = "uae_sound_stereo_separation"
    UAE_TOCCATA = "uae_toccata"
    UAE_WAITING_BLITS = "uae_waiting_blits"
    UAE_Z3CHIPMEM_SIZE = "uae_z3chipmem_size"
    UAE_Z3MAPPING = "uae_z3mapping"
    UAE_Z3MEM2_SIZE = "uae_z3mem2_size"
    UAE_Z3MEM_SIZE = "uae_z3mem_size"
    UAE_Z3MEM_START = "uae_z3mem_start"
    UAEGFX_CARD = "uaegfx_card"
    UAELOG = "uaelog"
    UAEM_WRITE_FLAGS = "uaem_write_flags"
    UAENATIVE_LIBRARY = "uaenative_library"
    UNSAFE_SAVE_STATES = "unsafe_save_states"
    VARIANT_NAME = "variant_name"
    VICE_AUDIO_DRIVER = "vice_audio_driver"
    VIDEO_FORMAT = "video_format"
    VIDEO_STANDARD = "video_standard"
    VIDEO_SYNC = "video_sync"
    VIDEO_SYNC_METHOD = "video_sync_method"
    VIEWPORT = "viewport"
    VOLUME = "volume"
    WARP_MODE = "warp_mode"
    WHDLOAD_BOOT_DIR = "whdload_boot_dir"
    WHDLOAD_MODEL = "whdload_model"
    WHDLOAD_PRELOAD = "whdload_preload"
    WHDLOAD_QUIT_KEY = "whdload_quit_key"
    WHDLOAD_SPLASH_DELAY = "whdload_splash_delay"
    WINDOW_BORDER = "window_border"
    WINDOW_HEIGHT = "window_height"
    WINDOW_HIDDEN = "window_hidden"
    WINDOW_MINIMIZED = "window_minimized"
    WINDOW_RESIZABLE = "window_resizable"
    WINDOW_WIDTH = "window_width"
    WINDOW_X = "window_x"
    WINDOW_Y = "window_y"
    WORKBENCH = "workbench"
    WRITABLE_FLOPPY_IMAGES = "writable_floppy_images"
    X_WHDLOAD_ARGS = "x_whdload_args"
    X_WHDLOAD_VERSION = "x_whdload_version"
    ZOOM = "zoom"
    ZORRO_III_MEMORY = "zorro_iii_memory"
    ZXS_DATABASE = "zxs_database"
    ZXS_DRIVER = "zxs_driver"
    ZXS_MODEL = "zxs_model"



import fsui
from launcher.option import Option
from launcher.ui.config.configpanel import ConfigPanel
from launcher.ui.floppiesgroup import FloppiesGroup
from launcher.ui.MediaListGroup import MediaListGroup
from launcher.ui.skin import Skin


class CDPanel(ConfigPanel):
    def __init__(self, parent):
        super().__init__(parent)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()
        self.drives_group = FloppiesGroup(self, 1, cd_mode=True)
        self.layout.add(self.drives_group, fill=True)
        self.layout.add_spacer(10)
        self.add_amiga_option(Option.CDROM_DRIVE_0_DELAY)
        self.layout.add_spacer(10)
        self.media_list_group = MediaListGroup(self, cd_mode=True)
        self.layout.add(self.media_list_group, expand=True, fill=True)



import os
import threading

from arcade.glui.state import State
from fsbc.settings import Settings
from fsui.qt import QCursor


class GameCenterRunner(object):
    def __init__(
        self, controller=None, platform=None, name=None, config=None, **_
    ):
        self.controller = controller
        self.platform_name = platform
        self.game_name = name
        self.config_name = config
        self.done = False
        self.status = ""
        self.error = ""
        self.backtrace = None

    def prepare(self):
        self.done = False
        threading.Thread(
            target=self._prepare_thread, name="GameRunnerPrepareThread"
        ).start()

    def _prepare_thread(self):
        try:
            self._do_prepare()
        except Exception as e:
            import traceback

            traceback.print_exc()
            self.error = repr(e)
            self.backtrace = traceback.format_exc()
        self.done = True
        self.signal()

    def _do_prepare(self):
        self.controller.prepare()
        self.controller.install()

    def configure(self):
        self.done = True

    #     self.status = ""
    #     self.error = ""
    #     self.done = False
    #     threading.Thread(target=self._configure_thread).start()
    #
    # def _configure_thread(self):
    #     try:
    #         self._do_configure()
    #     except Exception as e:
    #         import traceback
    #         traceback.print_exc()
    #         # signal error by setting status on done
    #         self.error = repr(e)
    #         self.backtrace = traceback.format_exc()
    #     self.done = True
    #     self.signal()
    #
    # def _do_configure(self):
    #     # self.controller.configure_game()
    #     pass

    def run(self):
        self.status = ""
        self.error = ""
        self.done = False

        cursor_position = os.environ.get("FSGS_RETURN_CURSOR_TO", "")
        try:
            x, y = cursor_position.split(",")
            x = int(x)
            y = int(y)
        except ValueError:
            pass
        else:
            # noinspection PyCallByClass,PyTypeChecker
            QCursor.setPos(x, y)

        # cursor_x = Settings.instance().get("__cursor_x")
        # cursor_y = Settings.instance().get("__cursor_y")
        # try:
        #     cursor_x = int(cursor_x)
        #     cursor_y = int(cursor_y)
        # except ValueError:
        #     pass
        # else:
        #     QCursor.setPos(cursor_x, cursor_y)

        threading.Thread(
            target=self._run_thread, name="GameRunnerRunThread"
        ).start()

    def abort(self):
        self.controller.abort()

    def _run_thread(self):
        try:
            self._do_run()
        except Exception as e:
            import traceback

            traceback.print_exc()
            self.error = repr(e)
            self.backtrace = traceback.format_exc()
        self.done = True
        self.signal()

    def _do_run(self):
        # if fs.windows:
        #     pass
        # elif fs.macosx:
        #     pass
        # else:
        #     # prevent flashing mouse cursor when launching programs
        #     # by moving cursor to the bottom right of the screen
        #     import gtk.gdk
        #     display = gtk.gdk.display_get_default()
        #     screen = display.get_default_screen()
        #     display.warp_pointer(screen, screen.get_width() - 1,
        #             screen.get_height() - 1)
        # size = get_fullscreen_size()
        # pygame.mouse.set_cursor((8, 1), (0, 0), (0,), (0,))
        # pygame.mouse.set_visible(True)
        # pygame.mouse.set_pos(size[0] - 1, size[1] - 0)

        state = State.get()
        try:
            print("state.game_running = True")
            state.game_running = True
            self.controller.run()
            self.controller.wait()
        finally:
            print("state.game_running = False")
            state.game_running = False

        self.controller.finish()

    def signal(self):
        # create a dummy event to wake up the main thread

        # event = pygame.event.Event(pygame.USEREVENT)
        # pygame.event.post(event)
        pass



from .spacer import Spacer


DEBUG = 0


class LayoutChild(object):
    def __init__(self):
        self.element = None
        self.spacing = 0
        self.expand = False
        self.fill = False
        self.valign = 0.5
        self.margin_left = 0
        self.margin_right = 0
        self.margin_top = 0
        self.margin_bottom = 0
        self.size = 0


class Layout(object):
    def __init__(self, padding):
        # self.min_size = (0, 0)
        self.position = (0, 0)
        self.size = (0, 0)
        self.padding_left = padding
        self.padding_right = padding
        self.padding_top = padding
        self.padding_bottom = padding
        self._skip = 0
        # self.origin = (0, 0)
        self.children = []
        self.min_width = 0
        self.min_height = 0

    def is_visible(self):
        return True

    def get_padding(self):
        return (
            self.padding_left,
            self.padding_top,
            self.padding_right,
            self.padding_bottom,
        )

    def set_padding(self, *amount):
        print("set_padding", amount)
        try:
            self.padding_left = amount[0]
            self.padding_top = amount[1]
            self.padding_right = amount[2]
            self.padding_bottom = amount[3]
        except IndexError:
            self.padding_left = amount[0]
            self.padding_right = amount[0]
            self.padding_top = amount[0]
            self.padding_bottom = amount[0]

    padding = property(get_padding, set_padding)

    def get_min_size(self):
        return self.get_min_width(), self.get_min_height()

    def insert(
        self,
        index,
        element,
        spacing=0,
        expand=False,
        fill=False,
        valign=0.5,
        margin=0,
        margin_left=None,
        margin_right=None,
        margin_top=None,
        margin_bottom=None,
    ):
        self.add(
            element,
            spacing=spacing,
            expand=expand,
            fill=fill,
            valign=valign,
            margin=margin,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            margin_bottom=margin_bottom,
            index=index,
        )

    # FIXME: Rename margin -> margins (+ alias), margin_top -> top, etc
    def add(
        self,
        element,
        spacing=0,
        expand=False,
        fill=False,
        valign=0.5,
        margin=0,
        margin_left=None,
        margin_right=None,
        margin_top=None,
        margin_bottom=None,
        index=None,
        left=None,
        right=None,
        top=None,
        bottom=None,
    ):
        """

        - By setting fill < 0, the height or width of the control will not
          be taken into account when calculating the layout height or width
          (depending on orientation). This makes it possible to center a
          larger control on a smaller layout

        - By setting expand < 0, the height or width of the control will not
          be taken into account as reserved space. It will still receive its
          share of expansion space, according to abs(expand) / sum(all expand).

        """
        if left is not None:
            margin_left = left
        if right is not None:
            margin_right = right
        if top is not None:
            margin_top = top
        if bottom is not None:
            margin_bottom = bottom
        child = LayoutChild()
        child.element = element
        child.spacing = spacing
        child.expand = expand
        child.fill = fill
        child.valign = valign
        if margin_left is not None:
            child.margin_left = margin_left
        else:
            child.margin_left = margin
        if margin_right is not None:
            child.margin_right = margin_right
        else:
            child.margin_right = margin
        if margin_top is not None:
            child.margin_top = margin_top
        else:
            child.margin_top = margin
        if margin_bottom is not None:
            child.margin_bottom = margin_bottom
        else:
            child.margin_bottom = margin
        if index is None:
            self.children.append(child)
        else:
            self.children.insert(index, child)

    def remove(self, element):
        for i, child in enumerate(self.children):
            if child.element == element:
                del self.children[i]
                return

    def add_spacer(self, size=0, size2=None, expand=False):
        self.add(Spacer(size, size2), expand=expand)

    def get_position(self):
        return self.position

    def set_position(self, position):
        self.position = position
        # self.origin = position
        # FIXME: avoid calling update after both set_position and set_size
        self.update()

    def set_size(self, size):
        if DEBUG:
            print("Layout.set_size", size)
        self.size = size
        self.update()

    def set_position_and_size(self, position, size):
        self.position = position
        self.size = size
        self.update()

    def update(self):
        pass


class LinearLayout(Layout):
    def __init__(self, vertical, padding=0):
        Layout.__init__(self, padding)
        self.vertical = vertical

    def update(self):
        available = self.size[self.vertical]
        if self.vertical:
            available = available - self.padding_top - self.padding_bottom
        else:
            available = available - self.padding_left - self.padding_right
        if DEBUG:
            print("update, available =", available)
        expanding = 0

        last_margin = 0

        for child in self.children:
            if not child.element.is_visible():
                continue

            child.min_size = [
                child.element.get_min_width(),
                child.element.get_min_height(),
            ]

            if child.expand < 0:
                child.size = 0
            else:
                child.size = child.min_size[self.vertical]

            # if child.fill < 0:
            #     child.min_size[not self.vertical] = 0

            available -= child.size
            expanding += abs(child.expand)

            if self.vertical:
                child._skip = max(last_margin, child.margin_top)
                available -= child._skip
                last_margin = child.margin_bottom
                # available = available - child.margin_top - child.margin_bottom
                # available = available - last_margin + max()
                # last_margin = child.margin_bottom
            else:
                child._skip = max(last_margin, child.margin_left)
                available -= child._skip
                last_margin = child.margin_right
        available -= last_margin

        if DEBUG:
            print("available", available, "expanding", expanding)
        if available > 0 and expanding > 0:
            if DEBUG:
                print("distributing extra pixels:", available)
            available2 = available
            for child in self.children:
                extra = int(available2 * (abs(child.expand) / expanding))
                if DEBUG:
                    print(child.expand, expanding, extra)
                child.size += extra
                available -= extra
            # some more pixels could be available due to rounding
            if available > 0:
                # print("distributing extra pixels:", available)
                for child in self.children:
                    if abs(child.expand):
                        child.size += 1
                        available -= 1
                        if available == 0:
                            break
        x = self.padding_left
        y = self.padding_top
        # self_height = self.size[1] - self.padding_top - self.padding_bottom
        fill_size = (
            self.size[0] - self.padding_left - self.padding_right,
            self.size[1] - self.padding_top - self.padding_bottom,
        )

        for child in self.children:
            if not child.element.is_visible():
                continue

            size = [child.min_size[0], child.min_size[1]]

            size[self.vertical] = child.size
            if child.fill > 0:
                size[not self.vertical] = fill_size[not self.vertical]
                if self.vertical:
                    size[0] -= child.margin_left + child.margin_right
                else:
                    size[1] -= child.margin_top + child.margin_bottom

            if DEBUG:
                print(child.element, size)

            self_pos = self.get_position()
            position = [self_pos[0] + x, self_pos[1] + y]

            position[self.vertical] += child._skip
            if self.vertical:
                position[0] += child.margin_left
            else:
                position[1] += child.margin_top

            if child.fill <= 0:
                # center child
                if self.vertical:
                    # position[0] += (self.size[0] - size[0]) // 2
                    pass
                else:
                    position[1] += round(
                        (
                            fill_size[1]
                            - child.margin_top
                            - child.margin_bottom
                            - size[1]
                        )
                        * child.valign
                    )

            child.element.set_position_and_size(position, size)

            if self.vertical:
                y += size[1] + child._skip
            else:
                x += size[0] + child._skip


class HorizontalLayout(LinearLayout):
    def __init__(self, padding=0):
        LinearLayout.__init__(self, False, padding=padding)

    def get_min_width(self):
        min_width = 0
        last_margin = 0
        for child in self.children:
            min_width += child.element.get_min_width()
            min_width += max(last_margin, child.margin_left)
            last_margin = child.margin_right
        min_width += last_margin
        min_width += self.padding_left + self.padding_right
        if min_width < self.min_width:
            return self.min_width
        return min_width

    def get_min_height(self):
        min_height = 0
        for child in self.children:
            h = child.element.get_min_height()
            if child.fill == -1:
                h = 0
            h += child.margin_top + child.margin_bottom
            if h > min_height:
                min_height = h
        min_height += self.padding_top + self.padding_bottom
        if min_height < self.min_height:
            return self.min_height
        return min_height


class VerticalLayout(LinearLayout):
    def __init__(self, padding=0):
        LinearLayout.__init__(self, True, padding=padding)

    def get_min_width(self):
        min_width = 0
        for child in self.children:
            if hasattr(child.element, "explicitly_hidden"):
                if child.element.explicitly_hidden():
                    continue
            w = child.element.get_min_width()
            w += child.margin_left + child.margin_right
            if w > min_width:
                min_width = w
        min_width += self.padding_left + self.padding_right
        if min_width < self.min_width:
            return self.min_width
        return min_width

    def get_min_height(self):
        min_height = 0
        last_margin = 0
        for child in self.children:
            if hasattr(child.element, "explicitly_hidden"):
                if child.element.explicitly_hidden():
                    continue
            min_height += child.element.get_min_height()
            min_height += max(last_margin, child.margin_top)
            last_margin = child.margin_bottom
        min_height += last_margin
        min_height += self.padding_top + self.padding_bottom
        if min_height < self.min_height:
            return self.min_height
        return min_height



import workspace
import workspace.ui
from workspace.util.help_button import HelpButton
from workspace.util.browse_button import BrowseButton


# FIXME: Extract ADF files as well


class ArchiveExtractorApplication(workspace.ui.Application):
    def __init__(self):
        super().__init__("archive_extractor")
        self.window = ArchiveExtractorWindow(self)
        self.window.show()

    def arguments(self, args):
        print("Archive Extractor player got args", args)
        self.window.set_playlist(args)


class ArchiveExtractorWindow(workspace.ui.Window):
    def __init__(self, parent):
        super().__init__(
            parent, "Archive Extractor", maximizable=False, menu=True
        )
        self.create_menu()
        col = workspace.ui.Column(self)
        col.add_spacer(540, 0)

        self.source_area = SourceArea(self)
        col.add(self.source_area)
        col.add(workspace.ui.TitleSeparator(self))

        # row = workspace.gui.Row()
        # col.add(row, margin=20)
        #
        # self.image_view = workspace.gui.ImageView(self, workspace.gui.Image(
        #     workspace.Stream(__name__, "data/package-x-generic.png")))
        # row.add(self.image_view, top=-10, bottom=-10, right=10)
        #
        # col2 = workspace.gui.Column()
        # row.add(col2, expand=True, fill=False, top=0)
        #
        # self.module_label = workspace.gui.Label(
        #     self, "No archive loaded", font="16px Medium Roboto")
        # col2.add(self.module_label)
        #
        # col.spacer(2)
        #
        # self.format_row = workspace.gui.Row()
        # col2.add(self.format_row)
        #
        # self.format_label = workspace.gui.Label(self, "SOURCE ARCHIVE")
        # # font = self.format_label.get_font()
        # # font.adjust_size(-1)
        # font = workspace.gui.Font("Roboto", 12)
        # self.format_label.set_font(font)
        # self.format_label.set_text_color(workspace.gui.Color(0x80, 0x80, 0x80))
        # self.format_row.add(self.format_label, expand=True)
        # # self.play_button = workspace.gui.Button(self, "Song")
        # # row.add(self.play_button)
        #
        # self.format_label_2 = workspace.gui.Label(self, "ZIP")
        # self.format_label_2.set_font(font)
        # self.format_label_2.set_text_color(workspace.gui.Color(0x80, 0x80, 0x80))
        # self.format_row.add(self.format_label_2, right=20)

        row = col.row(margin=20, bottom=10)
        row.add(workspace.ui.Label(self, "Destination folder" + ":"))
        row.expand()
        self.items_label = workspace.ui.Label(
            self, "4 root items will be created"
        )
        row.add(self.items_label)

        row = col.row(margin=20, top=0)
        self.dir_field = workspace.ui.TextField(self, "Ram:Temp4")
        row.add(self.dir_field, expand=True)
        self.browse_button = BrowseButton(self)
        row.add(self.browse_button, left=10)

        row = col.row(margin=20, top=0)
        # FIXME
        import fsui

        row.add(workspace.ui.Label(self, "Character set:"))
        self.charset_choice = fsui.Choice(
            self, ["ISO-8859-1", "UTF-8", "CP437"]
        )
        row.add(self.charset_choice, left=10)
        row.expand()
        row.add(workspace.ui.Label(self, "Create .uaem metadata:"))
        self.meta_choice = fsui.Choice(
            self, ["When needed", "Always", "Never"]
        )
        row.add(self.meta_choice, left=10)
        # self.meta_checkbox = fsui.CheckBox(
        #     self, "Create .uaem files when needed", True)
        # row.add(self.meta_checkbox, left=20)

        # row = workspace.gui.Row()
        # col.add(row, margin=20)
        #
        # self.image_view = workspace.gui.ImageView(self, workspace.gui.Image(
        #     workspace.Stream(__name__, "data/folder.png")))
        # row.add(self.image_view, right=10)
        #
        # col2 = workspace.gui.Column()
        # row.add(col2, expand=True, fill=False, top=0)
        #
        # self.module_label = workspace.gui.Label(
        #     self, "No destination selected", font="16px Medium Roboto")
        # col2.add(self.module_label)

        row = workspace.ui.Row()
        # FIXME
        row.add_spacer(0, 32)
        col.add(row, margin=20)

        button = HelpButton(self, "")
        row.add(button)

        button = workspace.ui.Button(self, "Extract")
        row.expand()
        row.add(button)

    def create_menu(self):
        menu = workspace.ui.Menu(self)
        menu.add_item("Load File", self.on_settings)
        menu.add_separator()
        menu.add_item("Settings", self.on_settings)
        menu.add_item("Reset to Defaults", self.on_settings)
        menu.add_separator()
        menu.add_item("About", self.on_about)
        self.set_menu(menu)

    def on_about(self):
        print("on_about")

    def on_settings(self):
        print("on_settings")


class SourceArea(workspace.ui.TitlePanel):
    def __init__(self, parent):
        super().__init__(parent)
        row = workspace.ui.Row(self, padding=20)

        self.image_view = workspace.ui.ImageView(
            self,
            workspace.ui.Image(
                workspace.Stream(__name__, "data/48/zip-file.png")
            ),
        )
        row.add(self.image_view, top=-10, bottom=-10, right=10)

        col = workspace.ui.Column()
        row.add(col, expand=True, fill=False, top=0)

        self.source_label = workspace.ui.Label(
            self, "No archive loaded", font="16px Medium Roboto"
        )
        col.add(self.source_label)

        col.spacer(2)
        self.format_row = workspace.ui.Row()
        col.add(self.format_row)

        # self.format_label = workspace.gui.Label(self, "ZIP ARCHIVE")
        self.format_label = workspace.ui.Label(self, "DROP FILE HERE TO LOAD")
        font = workspace.ui.Font("Roboto", 12)
        self.format_label.set_font(font)
        self.format_label.set_text_color(workspace.ui.Color(0x80, 0x80, 0x80))
        self.format_row.add(self.format_label, expand=True)
        # self.format_label_2 = workspace.gui.Label(self, "ISO-8859-1")
        # self.format_label_2.set_font(font)
        # self.format_label_2.set_text_color(workspace.gui.Color(0x80, 0x80, 0x80))
        # self.format_row.add(self.format_label_2)



import fsgs
import fsui
from launcher.option import Option
from launcher.settings.settings_page import SettingsPage


class ArcadeSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("fs-uae-arcade", "pkg:launcher")
        self.add_header(icon, "{} Arcade".format(fsgs.product))
        self.add_option(Option.ARCADE_FULLSCREEN)
        self.add_option(Option.ARCADE_THEME)
        self.add_option(Option.ARCADE_INITIAL_FAVORITES)



from fsgs.Database import Database
from fsgs.ogd.base import SynchronizerBase
from fsgs.res import gettext


class ListsSynchronizer(SynchronizerBase):
    def __init__(self, *args, **kwargs):
        SynchronizerBase.__init__(self, *args, **kwargs)

    def synchronize_list(self, database, list_uuid, list_info):
        self.remove_list(database, list_uuid)
        cursor = database.cursor()
        # cursor.execute("DELETE FROM game_list_game WHERE list_uuid = ?",
        #                (list_uuid,))
        doc = self.fetch_json("/api/list-sync/1/{0}".format(list_uuid))
        for game_info in doc["games"]:
            game_uuid = game_info["uuid"]
            variant_uuid = game_info.get("variant", None)
            position = game_info.get("position", None)
            cursor.execute(
                "INSERT INTO game_list_game (list_uuid, game_uuid,"
                "variant_uuid, position) VALUES (?, ?, ?, ?)",
                (list_uuid, game_uuid, variant_uuid, position),
            )
        cursor.execute(
            "INSERT INTO game_list (uuid, name, sync) VALUES (?, ?, ?)",
            (list_uuid, list_info["name"], list_info["sync"]),
        )

    def remove_list(self, database, list_uuid):
        cursor = database.cursor()
        cursor.execute("DELETE FROM game_list WHERE uuid = ?", (list_uuid,))
        cursor.execute(
            "DELETE FROM game_list_game WHERE list_uuid = ?", (list_uuid,)
        )

    def synchronize(self):
        if self.stop_check():
            return

        if "lists" not in self.context.meta:
            # Haven't looked up synchronization information from the server.
            return

        self.set_status(gettext("Updating game lists..."))

        database = Database.instance()
        cursor = database.cursor()
        cursor.execute("SELECT uuid, name, sync FROM game_list")
        # existing_lists = {}
        existing_syncs = {}
        for row in cursor:
            uuid, name, sync = row
            # existing_lists[uuid] = {
            #     "name": name,
            #     "sync": sync,
            # }
            existing_syncs[uuid] = sync
        # existing_syncs.sort()

        for list_uuid, list_info in self.context.meta["lists"].items():
            if list_info["sync"] == existing_syncs.get(list_uuid, None):
                print("[SYNC] List {} already synced".format(list_uuid))
            else:
                self.set_status(
                    gettext("Updating list '{0}'...".format(list_info["name"]))
                )
                self.synchronize_list(database, list_uuid, list_info)

        for existing_list_uuid in existing_syncs:
            for list_uuid in self.context.meta["lists"]:
                if list_uuid == existing_list_uuid:
                    break
            else:
                # this old list should be removed
                self.set_status(
                    gettext("Removing list {0}".format(existing_list_uuid))
                )
                self.remove_list(database, existing_list_uuid)

        database.commit()



from fsui import Image

from ...i18n import gettext
from ...launcher_config import LauncherConfig
from .StatusElement import StatusElement


class PlayersElement(StatusElement):
    def __init__(self, parent):
        StatusElement.__init__(self, parent)
        self.icon = Image("launcher:res/16x16/user.png")
        self.text = gettext("N/A")
        self.active = False

        self.players = ""

        LauncherConfig.add_listener(self)
        self.on_config("players", LauncherConfig.get("players"))

    def on_destroy(self):
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        if key == "players":
            if value != self.players:
                self.players = value
                if value:
                    self.text = value
                else:
                    self.text = gettext("N/A")
                self.active = bool(value)
                self.refresh()

    def get_min_width(self):
        return 96



import weakref
from fsui.qt import QMenu, QCursor, QPoint


class AutoCloseMenu(QMenu):
    def mousePressEvent(self, event):
        parent = self._parent()
        # noinspection PyArgumentList
        c = QCursor.pos()
        p = parent.widget().mapToGlobal(QPoint(0, 0))
        s = parent.get_size()
        if p.x() <= c.x() < p.x() + s[0] and p.y() <= c.y() < p.y() + s[1]:
            # We want to close the menu (and not re-open it) if we click
            # on the widget used to open the menu.
            # parent._ignore_next_left_down_event = True
            pass
        super().mousePressEvent(event)


class Menu:
    def __init__(self, implementation=QMenu):
        self.qmenu = implementation()
        # self._menu = wx.Menu()
        # self._ids = []
        # #self._functions = []
        pass

    def is_open(self):
        return self.qmenu.isVisible()

    def close(self):
        self.qmenu.close()

    def add_item(self, text, function=None, item_id=-1):
        text = text.replace("&", "&&")

        action = self.qmenu.addAction(text)
        if function is not None:
            action.triggered.connect(function)

        # if item_id == -1:
        #     item_id = wx.NewId()
        # self._ids.append(item_id)
        # #self._functions.append(function)
        # self._menu.Append(item_id, text)
        # def function_wrapper(event):
        #     # the reason we use CallAfter here is so control will return
        #     # to the application after having popped up a menu, so the
        #     # button opening a menu can draw its correct state when the
        #     # menu is closed, before for example a modal dialog is opened.
        #     wx.CallAfter(function)
        # self._menu.Bind(wx.EVT_MENU, function_wrapper, id=item_id)

    def add_about_item(self, text, function):
        self.add_item(text, function)

    def add_preferences_item(self, text, function):
        self.add_item(text, function)

    def add_separator(self):
        self.qmenu.addSeparator()

    def set_parent(self, parent):
        self._parent = weakref.ref(parent)
        self.qmenu._parent = weakref.ref(parent)


class PopupMenu(Menu):
    def __init__(self):
        super().__init__(implementation=AutoCloseMenu)



from fsui import Image, ImageButton


class IconButton(ImageButton):
    BUTTON_WIDTH = 40

    def __init__(self, parent, name):
        image = Image("launcher:res/" + name)
        ImageButton.__init__(self, parent, image)
        self.set_min_width(self.BUTTON_WIDTH)

    def set_icon_name(self, name):
        image = Image("launcher:res/" + name)
        self.set_image(image)



import fsui
from launcher.ui.Constants import Constants
from launcher.ui.skin import Skin
from launcher.ui.TabPanel import TabPanel


class TabButton(fsui.Panel):
    activated = fsui.Signal()

    TYPE_TAB = 0
    TYPE_BUTTON = 1

    STATE_NORMAL = 0
    STATE_SELECTED = 1

    def __init__(
        self,
        parent,
        icon,
        button_type=TYPE_TAB,
        left_padding=0,
        right_padding=0,
    ):
        fsui.Panel.__init__(self, parent, paintable=True)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()
        # self.set_background_color((0xdd, 0xdd, 0xdd))
        self.set_min_width(Constants.TAB_WIDTH + left_padding + right_padding)
        self.set_min_height(Constants.TAB_HEIGHT)
        self.group_id = 0
        self.icon = icon
        self.type = button_type
        self.state = self.STATE_NORMAL
        self.hover = False
        self.left_padding = left_padding
        self.right_padding = right_padding

    def select(self):
        self.get_parent().set_selected_tab(self)
        self.on_select()

    def on_paint(self):
        dc = self.create_dc()
        selected = self.state == self.STATE_SELECTED
        TabPanel.draw_background(
            self,
            dc,
            selected=selected,
            hover=self.hover,
            button_style=(self.type == self.TYPE_BUTTON),
        )
        # TabPanel.draw_border(self, dc)
        size = self.size()
        x = (
            self.left_padding
            + (
                size[0]
                - self.left_padding
                - self.right_padding
                - self.icon.size[0]
            )
            // 2
        )
        # subtracting two because of bottom border
        y = (size[1] - 2 - self.icon.size[1]) // 2
        dc.draw_image(self.icon, x, y)

    def check_hover(self):
        # FIXME: check if mouse is hovering, used for example after having
        # dismissed a popup menu
        self.hover = False
        self.refresh()

    def on_mouse_enter(self):
        self.hover = True
        self.refresh()

    def on_mouse_leave(self):
        self.hover = False
        self.refresh()

    def on_select(self):
        pass

    def on_left_down(self):
        if self.type == self.TYPE_TAB:
            self.select()

    def on_left_up(self):
        if self.type == self.TYPE_BUTTON:
            # FIXME: hack to avoid sticky hover due to mouse leave not
            # being sent if on_activate opens modal dialog
            # self.hover = False
            # self.refresh()
            self.on_activate()

    def on_activate(self):
        self.activated.emit()
